import abc
import warnings
from dataclasses import dataclass
from functools import reduce
from os import PathLike
from pathlib import Path
from string import Template
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from nilearn.surface import Mesh

from clinica.utils.stream import cprint

MISSING_TERM_ERROR_MSG = Template(
    "Term ${term} from the design matrix is not in the columns of the "
    "provided TSV file. Please make sure that there is no typo."
)


def _print_clusters(model: SLM, threshold: float) -> None:
    """This function prints the results related to total number
    of clusters, as well as the significative clusters.

    Parameters
    ----------
    model : brainstat.stats.SLM
        Fitted SLM model.

    threshold : float
        Cluster defining threshold.
    """
    cprint("#" * 40)
    cprint("After correction (Cluster-wise Correction for Multiple Comparisons): ")
    df = model.P["clus"][0]
    cprint(df)
    cprint(f"Clusters found: {len(df)}")
    cprint(
        f"Significative clusters (after correction): {len(df[df['P'] <= threshold])}"
    )


def _check_column_in_df(df: pd.DataFrame, column: str) -> None:
    """Checks if the provided column name is in the provided DataFrame.
    Raises a ValueError if not.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to analyze.

    column : str
        Name of the column to check.
    """
    if column not in df.columns:
        raise ValueError(MISSING_TERM_ERROR_MSG.safe_substitute(term=column))


def _categorical_column(df: pd.DataFrame, column: str) -> bool:
    """Returns `True` if the column is categorical and `False` otherwise.

    Parameters
    ----------
    df : pd.DataFrame
        The DataFrame to analyze.

    column : str
        The name of the column to check.

    Returns
    -------
    bool :
        `True` if the column contains categorical values, `False` otherwise.
    """
    return not df[column].dtype.name.startswith("float")


def _build_model(design_matrix: str, df: pd.DataFrame) -> FixedEffect:
    """Build a brainstat model from the design matrix in
    string format.

    This function assumes that the design matrix is formatted
    in the following way:

        1 + factor_1 + factor_2 + ...

    Or:

        factor_1 + factor_2 + ...

    in the latter case the intercept will be added automatically.

    Parameters
    ----------
    design_matrix : str
        Design matrix specified as a string.

    df : pd.DataFrame
        Subjects DataFrame.

    Returns
    -------
    model : FixedEffect
        BrainStats model.
    """
    if len(design_matrix) == 0:
        raise ValueError("Design matrix cannot be empty.")
    if "+" in design_matrix:
        terms = [_.strip() for _ in design_matrix.split("+")]
    else:
        terms = [design_matrix.strip()]
    model = []
    for term in terms:
        # Intercept is automatically included in brainstat
        if term == "1":
            continue
        # Handles the interaction effects
        if "*" in term:
            sub_terms = [_.strip() for _ in term.split("*")]
            model_term = reduce(
                lambda x, y: x * y, [_build_model_term(_, df) for _ in sub_terms]
            )
        else:
            model_term = _build_model_term(term, df)
        model.append(model_term)
    return reduce(lambda x, y: x + y, model)


def _build_model_term(
    term: str,
    df: pd.DataFrame,
    add_intercept: Optional[bool] = True,
) -> FixedEffect:
    """Builds a BrainStats model term from the subjects
    DataFrame and a column name.

    Parameters
    ----------
    term : str
        The name of the column of the DataFrame to be used.

    df : pd.DataFrame
        The subjects DataFrame.

    add_intercept : bool
        If `True`, adds an intercept term.

    Returns
    -------
    FixedEffect :
        BrainStats model term.
    """
    return FixedEffect(df[term], add_intercept=add_intercept)


class GLM:
    """This class implements the functionalities common to all GLM models
    used in the Clinica SurfaceStatistics pipeline.

    Attributes
    ----------
    design : str
        The design matrix specified in string format.
        If this contains a "*", it will be interpreted as an interaction effect.

    df : pd.DataFrame
        The subjects DataFrame.

    feature_label : str
        The label used for building output filenames.

    contrast : str
        The contrast specified in string format.

    fwhm : int, optional
        The smoothing FWHM. This is used in the output file names.
        Default=20.

    threshold_uncorrected_pvalue : float, optional
        The threshold to be used with uncorrected P-values. Default=0.001.

    threshold_corrected_pvalue : float, optional
        The threshold to be used with corrected P-values. Default=0.05.

    cluster_threshold : float, optional
        The threshold to be used to declare clusters as significant. Default=0.001.
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        self._two_tailed = False
        self._correction = ["fdr", "rft"]
        self.df = df
        self.feature_label = feature_label
        self.fwhm = fwhm
        self.threshold_uncorrected_pvalue = threshold_uncorrected_pvalue
        self.threshold_corrected_pvalue = threshold_corrected_pvalue
        self.cluster_threshold = cluster_threshold
        self.results_ = None
        self.slm_models_ = None
        self.contrasts = dict()
        self.filenames = dict()
        self.model = _build_model(design, df)
        self.build_contrasts(contrast)

    @property
    def contrast_names(self) -> List[str]:
        if self.contrasts is not None:
            return list(self.contrasts.keys())
        return list()

    @property
    def results(self):
        if self._is_fitted():
            return self.results_

    @abc.abstractmethod
    def build_contrasts(self, contrast: str):
        """Build the contrasts from the provided contrast in string format.

        .. note::
            This method needs to be implemented in subclasses.

        Parameters
        ----------
        contrast : str
            Contrast in string format.
        """
        pass

    @abc.abstractmethod
    def filename_root(self, contrast: str):
        """Returns the output file name root for the provided contrast.

        .. note::
            This method needs to be implemented in subclasses.

        Parameters
        ----------
        contrast : str
            Contrast for which to get the output filename.
        """
        pass

    def _is_fitted(self) -> bool:
        return self.results_ is not None

    def fit(
        self, data: np.ndarray, surface: Dict, mask: Optional[np.ndarray] = None
    ) -> None:
        """Fit the GLM model instance.

        Parameters
        ----------
        data : np.ndarray
            The data on which to fit the GLM model.

        surface : dict
            The Brainstat surface on which to fit the GLM model.

        mask : np.ndarray, optional
            The mask to be used to mask the data. Default=None.
        """
        if mask is None:
            mask = data[0, :] > 0
        self.results_ = dict()
        self.slm_models_ = dict()
        for contrast_name, contrast in self.contrasts.items():
            slm_model = SLM(
                self.model,
                contrast=contrast,
                surf=surface,
                mask=mask,
                two_tailed=self._two_tailed,
                correction=self._correction,
                cluster_threshold=self.cluster_threshold,
            )
            cprint(
                msg=f"Fitting the GLM model with contrast {contrast_name}...",
                lvl="info",
            )
            slm_model.fit(data)
            _print_clusters(slm_model, self.threshold_corrected_pvalue)
            self.results_[contrast_name] = StatisticsResults.from_slm_model(
                slm_model,
                mask,
                self.threshold_uncorrected_pvalue,
                self.threshold_corrected_pvalue,
            )
            self.slm_models_[contrast_name] = slm_model

    def save_results(self, output_dir: PathLike, method: Union[str, List[str]]) -> None:
        """Save results to the provided output directory.

        Parameters
        ----------
        output_dir : PathLike
            The output directory in which to write the results.

        method : str or List[str]
            The method(s) to write the results.
        """
        if not self._is_fitted():
            raise ValueError(
                "GLM model needs to be fitted before accessing the results."
            )
        if isinstance(method, str):
            method = [method]
        for contrast, result in self.results_.items():
            result_serializer = StatisticsResultsSerializer(
                Path(output_dir) / Path(self.filename_root(contrast))
            )
            for meth in method:
                result_serializer.save(result, meth)

    def plot_results(
        self,
        output_dir: PathLike,
        method: Union[str, List[str]],
        mesh: Mesh,
    ) -> None:
        """Plot results to the provided directory.

        Parameters
        ----------
        output_dir : PathLike
            The output directory in which to write the plot files.

        method : str or List[str]
            The method(s) to make the plots.

        mesh : nilearn.surface.Mesh
            The mesh on which to plot the result data.
        """
        if not self._is_fitted():
            raise ValueError(
                "GLM model needs to be fitted before accessing the results."
            )
        if isinstance(method, str):
            method = [method]
        for contrast, result in self.results_.items():
            plotter = StatisticsResultsPlotter(
                Path(output_dir) / Path(self.filename_root(contrast)), mesh
            )
            for meth in method:
                plotter.plot(result, meth)


class CorrelationGLM(GLM):
    """Class implementing the correlation type GLM model.

    Attributes
    ----------
    See documentation for `GLM` class.

    group_label : str, optinal
        The label to use for group GLM models. Default=None.
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str],
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        self.with_interaction = False
        self.absolute_contrast_name = None
        self.contrast_sign = None
        self.group_label = group_label
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )

    def build_contrasts(self, contrast: str):
        """Build the contrast from the string specification.

        Parameters
        ----------
        contrast : str
            The contrast to build.
        """
        absolute_contrast_name = contrast
        contrast_sign = "positive"
        if contrast.startswith("-"):
            absolute_contrast_name = contrast[1:].lstrip()
            contrast_sign = "negative"
        built_contrast = self.df[absolute_contrast_name]
        if contrast_sign == "negative":
            built_contrast *= -1
        self.contrasts[contrast] = built_contrast
        self.absolute_contrast_name = absolute_contrast_name
        self.contrast_sign = contrast_sign

    def filename_root(self, contrast: str):
        """Build the filename root part from class attributes and provided contrast.

        Parameters
        ----------
        contrast : str
            The contrast to use for building the filename.
        """
        if contrast not in self.contrasts:
            raise ValueError(f"Unknown contrast {contrast}.")
        return (
            f"group-{self.group_label}_correlation-{self.absolute_contrast_name}"
            f"-{self.contrast_sign}_measure-{self.feature_label}_fwhm-{self.fwhm}"
        )


class GroupGLM(GLM):
    """Class implementing group GLM models.

    Attributes
    ----------
    See documentation for `GLM` class.

    group_label : str, optional
        The Label to use for group GLM models. Default="group".
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str] = "group",
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        self.with_interaction = False
        self.group_label = group_label
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )

    def build_contrasts(self, contrast: str):
        """Build the contrast from the string specification.

        Parameters
        ----------
        contrast : str
            The contrast to build.
        """
        _check_column_in_df(self.df, contrast)
        if not _categorical_column(self.df, contrast):
            raise ValueError(
                "Contrast should refer to a categorical variable for group comparison. "
                "Please select 'correlation' for 'glm_type' otherwise."
            )
        group_values = np.unique(self.df[contrast])
        for contrast_type, (i, j) in zip(["positive", "negative"], [(0, 1), (1, 0)]):
            contrast_name = f"{group_values[j]}-lt-{group_values[i]}"
            self.contrasts[contrast_name] = (
                self.df[contrast] == group_values[i]
            ).astype(int) - (self.df[contrast] == group_values[j]).astype(int)

    def filename_root(self, contrast: str):
        """Build the filename root part from class attributes and provided contrast.

        Parameters
        ----------
        contrast : str
            The contrast to use for building the filename.
        """
        if contrast not in self.contrasts:
            raise ValueError(f"Unknown contrast {contrast}.")
        return f"group-{self.group_label}_{contrast}_measure-{self.feature_label}_fwhm-{self.fwhm}"


class GroupGLMWithInteraction(GroupGLM):
    """This class implements a GLM model for group comparison with
    interaction effects.

    Attributes
    ----------
    See attributes of parent class `GroupGLM`.
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str] = "group",
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            group_label,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )
        self.with_interaction = True
        warnings.warn(
            "You included interaction as covariate in your model, "
            "please carefully check the format of your tsv files."
        )

    def build_contrasts(self, contrast: str):
        """Build the contrast from the string specification.

        Parameters
        ----------
        contrast : str
            The contrast to build.
        """
        contrast_elements = [_.strip() for _ in contrast.split("*")]
        for contrast_element in contrast_elements:
            _check_column_in_df(self.df, contrast_element)
        categorical = [_categorical_column(self.df, _) for _ in contrast_elements]
        if len(contrast_elements) != 2 or sum(categorical) != 1:
            raise ValueError(
                "The contrast must be an interaction between one continuous "
                "variable and one categorical variable. Your contrast contains "
                f"the following variables : {contrast_elements}"
            )
        idx = 0 if categorical[0] else 1
        categorical_contrast = contrast_elements[idx]
        continue_contrast = contrast_elements[(idx + 1) % 2]
        group_values = np.unique(self.df[categorical_contrast])
        built_contrast = self.df[continue_contrast].where(
            self.df[categorical_contrast] == group_values[0], 0
        ) - self.df[continue_contrast].where(
            self.df[categorical_contrast] == group_values[1], 0
        )
        self.contrasts[contrast] = built_contrast

    def filename_root(self, contrast: str):
        """Build the filename root part from class attributes and provided contrast.

        Parameters
        ----------
        contrast : str
            The contrast to use for building the filename.
        """
        if contrast not in self.contrasts:
            raise ValueError(f"Unknown contrast {contrast}.")
        return f"interaction-{contrast}_measure-{self.feature_label}_fwhm-{self.fwhm}"


def create_glm_model(
    glm_type: str,
    design: str,
    df: pd.DataFrame,
    contrast: str,
    feature_label: str,
    group_label: Optional[str] = "group",
    fwhm: Optional[int] = 20,
    threshold_uncorrected_pvalue: Optional[float] = 0.001,
    threshold_corrected_pvalue: Optional[float] = 0.05,
    cluster_threshold: Optional[float] = 0.001,
) -> GLM:
    """Factory method for building a GLM model instance corresponding to the
    provided type and design matrix.

    Parameters
    ----------
    glm_type : str
        The type of GLM to be created. Either "correlation" or "group_comparison".

    design : str
        The design matrix specified in string format.
        If this contains a "*", it will be interpreted as an interaction effect.

    df : pd.DataFrame
        The subjects DataFrame.

    contrast : str
        The contrast specified in string format.

    feature_label : str
        The label used for building output filenames.

    group_label : str, optional
        The label to use for group GLM models. Default="group".

    fwhm : int, optional
        The smoothing FWHM. This is used in the output file names.
        Default=20.

    threshold_uncorrected_pvalue : float, optional
        The threshold to be used with uncorrected P-values. Default=0.001.

    threshold_corrected_pvalue : float, optional
        The threshold to be used with corrected P-values. Default=0.05.

    cluster_threshold : float, optional
        The threshold to be used to declare clusters as significant. Default=0.001.

    Returns
    -------
    model : GLM
        An instance of the `GLM` class.

    Raises
    ------
    ValueError
        If the glm_type is not supported.
    """
    cprint(
        msg=f"The GLM model is: {design} and the GLM type is: {glm_type}",
        lvl="info",
    )
    params = {
        "group_label": group_label,
        "fwhm": fwhm,
        "threshold_uncorrected_pvalue": threshold_uncorrected_pvalue,
        "threshold_corrected_pvalue": threshold_corrected_pvalue,
        "cluster_threshold": cluster_threshold,
    }
    if glm_type == "correlation":
        return CorrelationGLM(design, df, feature_label, contrast, **params)
    elif glm_type == "group_comparison":
        if "*" in design:
            return GroupGLMWithInteraction(
                design, df, feature_label, contrast, **params
            )
        return GroupGLM(design, df, feature_label, contrast, **params)
    raise ValueError(
        f"create_glm_model received an unknown GLM type: {glm_type}."
        f"Only 'correlation' and 'group_comparison' are supported."
    )


def _convert_arrays_to_lists(data: dict) -> dict:
    """If the input dictionary contains numpy arrays, this function will
    cast them to lists and return the same dictionary with the lists instead
    of the numpy arrays.

    Parameters
    ----------
    data : dict
        The dictionary to clean.

    Returns
    -------
    new_data : dict
        The dictionary with arrays casted to lists.
    """
    new_data = dict()
    for k, v in data.items():
        if isinstance(v, dict):
            new_data[k] = _convert_arrays_to_lists(v)
        elif isinstance(v, np.ndarray):
            new_data[k] = v.tolist()
        else:
            new_data[k] = v
    return new_data


class Results:
    """Common class for GLM results."""

    def to_dict(self) -> dict:
        """Returns the `Results` instance in dict format.

        Private attributes and all methods are not returned.

        This function does not perform any casting.

        Returns
        -------
        data : dict
            Resulting dictionary.
        """
        import inspect

        data = dict()
        for attribute in inspect.getmembers(self):
            name, value = attribute
            if not name.startswith("_"):
                if not inspect.ismethod(value):
                    if hasattr(value, "to_dict"):
                        data[name] = value.to_dict()
                    else:
                        data[name] = value
        return data

    def to_json(self, indent: Optional[int] = 4) -> str:
        """Returns the json of the `Results` instance.

        Parameters
        ----------
        indent : int, optional
            Indent to use. Default=4.

        Returns
        -------
        str :
            The JSON dumps of the results.
        """
        import json

        return json.dumps(_convert_arrays_to_lists(self.to_dict()), indent=indent)


@dataclass
class PValueResults(Results):
    """This class implements a container for raw (uncorrected)
    P-value results obtained with a GLM model.

    Attributes
    ----------
    pvalues : np.ndarray
        Array of uncorrected P-values.

    mask : np.ndarray
        The binary mask.

    threshold : float
        The threshold used.
    """

    pvalues: np.ndarray
    mask: np.ndarray
    threshold: float

    @property
    def thresh(self):
        """For compatibility with previous Matlab implementation."""
        return self.threshold

    @property
    def P(self):
        """For compatibility with previous Matlab implementation."""
        return self.pvalues

    @classmethod
    def from_t_statistics(
        cls,
        tstats: np.ndarray,
        df: pd.DataFrame,
        mask: np.ndarray,
        threshold: float,
    ):
        """Instantiate the class from an array of T-statistics.

        Parameters
        ----------
        tstats : np.ndarray
            Array of T-statistics.

        df : pd.DataFrame
            The subjects DataFrame.

        mask : np.ndarray
            The binary mask.

        threshold : float
            The threshold to be used.
        """
        from scipy.stats import t

        return cls(1 - t.cdf(tstats, df), mask, threshold)


@dataclass
class CorrectedPValueResults(PValueResults):
    """This class implements a container for corrected P-value
    results obtained with a GLM model.

    Attributes
    ----------
    cluster_pvalues : np.ndarray
        The cluster P-values.
    """

    cluster_pvalues: np.ndarray

    @property
    def C(self):
        """For compatibility with previous Matlab implementation."""
        return self.cluster_pvalues


@dataclass
class StatisticsResults(Results):
    """This class implements a container for results obtained with
    the GLM model classes. It holds information relative to a GLM
    run with one specific contrast.

    Attributes
    ----------
    coefficients : np.ndarray
        The beta coefficients of the fitted GLM model.

    tstats : np.ndarray
        The corresponding T-statistics.

    uncorrected_p_value : PValueResults
        The corresponding uncorrected p values, stored in a `PValueResults` instance.

    fdr : np.ndarray
        The corresponding False Discovery Rate.

    corrected_p_value : CorrectedPValueResults
        The corresponding corrected p values, stored in a `CorrectedPValueResults` instance.
    """

    coefficients: np.ndarray
    tstats: np.ndarray
    uncorrected_p_values: PValueResults
    fdr: np.ndarray
    corrected_p_values: CorrectedPValueResults

    @property
    def TStatistics(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.tstats

    @property
    def uncorrectedPValue(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.uncorrected_p_values

    @property
    def correctedPValue(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.corrected_p_values

    @property
    def FDR(self):
        """Needed for compatibility with previous implementation in Matlab."""
        return self.fdr

    @classmethod
    def from_slm_model(
        cls,
        model: SLM,
        mask: np.ndarray,
        threshold_uncorrected_p_value: float,
        threshold_corrected_p_value: float,
    ):
        """Instanciate from a SLM model.

        Parameters
        ----------
        model : brainstat.stats.SLM
            SLM model instance to use.

        mask : np.ndarray
            The binary mask to use.

        threshold_uncorrected_p_value : float
            The threshold to use with uncorrected P-values.

        threshold_corrected_p_value : float
            The threshold to use with corrected P-values.
        """
        idx = np.argwhere(np.isnan(model.t))
        corrected_pvals = model.P["pval"]["P"]
        corrected_pvals[idx] = 1.0
        tstats = np.nan_to_num(model.t)
        uncorrected_p_values = PValueResults.from_t_statistics(
            tstats,
            model.df,
            mask,
            threshold_uncorrected_p_value,
        )
        corrected_p_values = CorrectedPValueResults(
            corrected_pvals,
            model.P["pval"]["C"],
            mask,
            threshold_corrected_p_value,
        )
        return cls(
            np.nan_to_num(model.coef),
            tstats,
            uncorrected_p_values,
            model.Q,
            corrected_p_values,
        )


class StatisticsResultsPlotter:
    """Class responsible to plotting results of GLM fit.

    Attributes
    ----------
    output_file : PathLike
        Path to the output file.

    mesh : nilearn.surface.Mesh
        The mesh to be used for plotting results.
    """

    def __init__(self, output_file: PathLike, mesh: Mesh):
        self.output_file = output_file
        self.mesh = mesh
        self.plotting_extension = ".png"
        self.no_plot = {"coefficients"}  # Elements which should not be plotted

    def plot(self, result: StatisticsResults, method: str) -> None:
        """Plot the results.

        Parameters
        ----------
        result : StatisticsResults
            The results to be plotted.

        method : str
            The plotting method to use.
        """
        plotter = self._get_plotter(method)
        plotter(result)

    def _get_plotter(self, method: str) -> Callable[[StatisticsResults], None]:
        """Returns the plotting method from its name.

        Parameters
        ----------
        method : str
            Name of the plotting method to use.

        Returns
        -------
        Callable :
            Plotting method.
        """
        if method == "nilearn_plot_surf_stat_map":
            return self._plot_stat_maps
        else:
            raise NotImplementedError(f"Plotting method {method} is not implemented.")

    def _plot_stat_maps(self, result: StatisticsResults) -> None:
        """Wrapper around the `nilearn.plotting.plot_surf_stat_map` method.

        Parameters
        ----------
        result : StatisticsResults
            The results to plot.
        """
        from nilearn.plotting import plot_surf_stat_map

        for name, res in result.to_dict(jsonable=False).items():
            if name not in self.no_plot:
                texture = res
                threshold = None
                plot_filename = (
                    str(self.output_file) + "_" + name + self.plotting_extension
                )
                if isinstance(res, dict):
                    texture = res["P"]
                    threshold = res["thresh"]
                cprint(msg=f"Saving plot to {plot_filename}", lvl="info")
                plot_surf_stat_map(
                    self.mesh,
                    texture,
                    threshold=threshold,
                    output_file=plot_filename,
                    title=name,
                )


class StatisticsResultsSerializer:
    """This class is responsible for writing instances of `StatisticsResults`
    to disk through different methods.

    Attributes
    ----------
    output_file : PathLike
        Path and filename root to be used.
    """

    def __init__(self, output_file: PathLike):
        self.output_file = output_file
        self.json_extension = "_results.json"
        self.json_indent = 4
        self.mat_extension = ".mat"

    def save(self, result: StatisticsResults, method: str) -> None:
        """Save provided `StatisticsResults` to disk with provided method.

        Parameters
        ----------
        result : StatisticsResults
            Results to be saved.

        method : str
            Name of the saving method to use.
        """
        writer = self._get_writer(method)
        writer(result)

    def _get_writer(self, method: str) -> Callable[[StatisticsResults], None]:
        """Returns a writter method from its name.

        Parameters
        ----------
        method : str
            The name of the writting method to use.

        Returns
        -------
        Callable :
            The writting method.
        """
        if method.lower() == "json":
            return self._write_to_json
        elif method.lower() == "mat":
            return self._write_to_mat
        else:
            raise NotImplementedError(
                f"Serializing method {method} is not implemented."
            )

    def _write_to_json(self, results: StatisticsResults) -> None:
        """Write the provided `StatisticsResults` to JSON format.

        Parameters
        ----------
        results : StatisticsResults
            The results to write to disk in JSON format.
        """
        import json
        import os

        out_json_file = Path(str(self.output_file) + self.json_extension)
        if not os.path.exists(out_json_file.parents[0]):
            os.makedirs(out_json_file.parents[0])
        cprint(
            msg=f"Writing results to JSON in {out_json_file}...",
            lvl="info",
        )
        with open(out_json_file, "w") as fp:
            json.dump(results.to_json(indent=self.json_indent), fp)

    def _write_to_mat(self, results: StatisticsResults) -> None:
        """Write the provided `StatisticsResults` to MAT format.

        Parameters
        ----------
        results : StatisticsResults
            The results to write to disk in MAT format.
        """
        from scipy.io import savemat

        # These labels are used for compatibility with the previous
        # MATLAB implementation of the Statistics Surface Pipeline
        # of Clinica.
        struct_labels = {
            "coefficients": "coef",
            "TStatistics": "tvaluewithmask",
            "uncorrectedPValue": "uncorrectedpvaluesstruct",
            "correctedPValue": "correctedpvaluesstruct",
            "FDR": "FDR",
        }
        for name, res in results.to_dict().items():
            if name in struct_labels:
                mat_filename = str(self.output_file) + "_" + name + self.mat_extension
                cprint(
                    msg=f"Writing {name} results to MAT in  {mat_filename}",
                    lvl="info",
                )
                savemat(mat_filename, {struct_labels[name]: res})



from typing import List, Optional

from clinica.iotools.abstract_converter import Converter


def get_bids_subjs_info(
    clinical_data_dir: str,
    out_path: str,
    subjects_list_path: Optional[str] = None,
):
    from os import path

    from pandas import read_csv

    # Read optional list of participants.
    subjects_list = (
        set([line.rstrip("\n") for line in open(subjects_list_path)])
        if subjects_list_path
        else None
    )

    # Load all participants from ADNIMERGE.
    adni_merge_path = path.join(clinical_data_dir, "ADNIMERGE.csv")
    participants = set(
        read_csv(adni_merge_path, sep=",", usecols=["PTID"], squeeze=True).unique()
    )

    # Filter participants if requested.
    participants = sorted(
        participants & subjects_list if subjects_list else participants
    )

    # Compute their corresponding BIDS IDs and paths.
    bids_ids = [f"sub-ADNI{p.replace('_', '')}" for p in participants]
    bids_paths = [path.join(out_path, bids_id) for bids_id in bids_ids]

    return bids_ids, bids_paths


class AdniToBids(Converter):
    @classmethod
    def get_modalities_supported(cls) -> List[str]:
        """Return a list of modalities supported.

        Returns: a list containing the modalities supported by the converter
        (T1, PET_FDG, PET_AMYLOID, PET_TAU, DWI, FLAIR, fMRI)
        """
        return ["T1", "PET_FDG", "PET_AMYLOID", "PET_TAU", "DWI", "FLAIR", "fMRI"]

    @classmethod
    def check_adni_dependencies(cls) -> None:
        """Check the dependencies of ADNI converter."""
        from clinica.utils.check_dependency import check_dcm2niix

        check_dcm2niix()

    def convert_clinical_data(
        self,
        clinical_data_dir: str,
        out_path: str,
        clinical_data_only: bool = False,
        subjects_list_path: Optional[str] = None,
        xml_path: Optional[str] = None,
    ):
        """Convert the clinical data of ADNI specified into the file clinical_specifications_adni.xlsx.

        Args:
            clinical_data_dir:  path to the clinical data directory
            out_path: path to the BIDS directory
            clinical_data_only: process clinical data only
            subjects_list_path: restrict processing to this manifest of subjects
            xml_path: path to the XML metadata files
        """
        import os
        from os import path

        import clinica.iotools.bids_utils as bids
        import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
        from clinica.utils.stream import cprint

        from .adni_json import create_json_metadata

        clinic_specs_path = path.join(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
            "data",
            "clinical_specifications_adni",
        )
        try:
            os.path.exists(out_path)
        except IOError:
            print("BIDS folder not found.")
            raise

        conversion_path = path.join(out_path, "conversion_info")

        if clinical_data_only:
            bids_ids, bids_subjs_paths = get_bids_subjs_info(
                clinical_data_dir=clinical_data_dir,
                out_path=out_path,
                subjects_list_path=subjects_list_path,
            )
        else:
            bids_ids = bids.get_bids_subjs_list(out_path)
            bids_subjs_paths = bids.get_bids_subjs_paths(out_path)

        # -- Creation of modality agnostic files --
        cprint("Creating modality agnostic files...")
        readme_data = {
            "link": "http://adni.loni.usc.edu",
            "desc": (
                "ADNI is a global research effort that actively supports the investigation and development of "
                "treatments that slow or stop the progression of Alzheimer's disease (AD).This multisite, longitudinal "
                "study assesses clinical, imaging, genetic and biospecimen biomarkers through the process of normal "
                "aging to mild cognitive impairment (MCI) and AD dementia.With established, standardized methods for "
                "imaging and biomarker collection and analysis, ADNI facilitates a way for scientists to conduct "
                "cohesive research and share compatible data with other researchers around the world."
            ),
        }
        bids.write_modality_agnostic_files(
            study_name="ADNI", readme_data=readme_data, bids_dir=out_path
        )

        # -- Creation of participant.tsv --
        cprint("Creating participants.tsv...")
        participants_df = bids.create_participants_df(
            "ADNI", clinic_specs_path, clinical_data_dir, bids_ids
        )

        # Replace the original values with the standard defined by the AramisTeam
        participants_df["sex"] = participants_df["sex"].replace("Male", "M")
        participants_df["sex"] = participants_df["sex"].replace("Female", "F")

        # Correction of diagnosis_sc for ADNI3 participants
        participants_df = adni_utils.correct_diagnosis_sc_adni3(
            clinical_data_dir, participants_df
        )

        participants_df.to_csv(
            path.join(out_path, "participants.tsv"),
            sep="\t",
            index=False,
            encoding="utf-8",
        )

        # -- Creation of sessions.tsv --
        cprint("Creating sessions files...")
        adni_utils.create_adni_sessions_dict(
            bids_ids, clinic_specs_path, clinical_data_dir, bids_subjs_paths
        )

        # -- Creation of scans files --
        if os.path.exists(conversion_path):
            cprint("Creating scans files...")
            adni_utils.create_adni_scans_files(conversion_path, bids_subjs_paths)

        if xml_path is not None:
            if os.path.exists(xml_path):
                create_json_metadata(bids_subjs_paths, bids_ids, xml_path)
            else:
                cprint(
                    msg=(
                        f"Clinica was unable to find {xml_path}, "
                        "skipping xml metadata extraction."
                    ),
                    lvl="warning",
                )

    def convert_images(
        self,
        source_dir,
        clinical_dir,
        dest_dir,
        subjs_list_path=None,
        modalities=None,
        force_new_extraction=False,
    ):
        """Convert the images of ADNI.

        Args:
            source_dir: path to the ADNI directory
            clinical_dir: path to the clinical data directory
            dest_dir: path to the BIDS directory
            subjs_list_path: list of subjects to process
            modalities: modalities to convert (T1, PET_FDG, PET_AMYLOID, PET_TAU, DWI, FLAIR, fMRI)
            force_new_extraction: if given pre-existing images in the BIDS directory will be erased and extracted again.
        """
        import os
        from copy import copy
        from os import path

        import pandas as pd

        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_av45_fbb_pet as adni_av45_fbb
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_dwi as adni_dwi
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_fdg_pet as adni_fdg
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_flair as adni_flair
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_fmri as adni_fmri
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_pib_pet as adni_pib
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_t1 as adni_t1
        import clinica.iotools.converters.adni_to_bids.adni_modalities.adni_tau_pet as adni_tau
        from clinica.utils.stream import cprint

        modalities = modalities or self.get_modalities_supported()

        adni_merge_path = path.join(clinical_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path)

        # Load a file with subjects list or compute all the subjects
        if subjs_list_path is not None:
            cprint("Loading a subjects lists provided by the user...")
            subjs_list = [line.rstrip("\n") for line in open(subjs_list_path)]
            subjs_list_copy = copy(subjs_list)

            # Check that there are no errors in subjs_list given by the user
            for subj in subjs_list_copy:
                adnimerge_subj = adni_merge[adni_merge.PTID == subj]
                if len(adnimerge_subj) == 0:
                    cprint(
                        msg=f"Subject with PTID {subj} does not exist. Please check your subjects list.",
                        lvl="warning",
                    )
                    subjs_list.remove(subj)
            del subjs_list_copy

        else:
            cprint("Using all the subjects contained into the ADNIMERGE.csv file...")
            subjs_list = list(adni_merge["PTID"].unique())

        # Create the output folder if is not already existing
        os.makedirs(dest_dir, exist_ok=True)
        os.makedirs(path.join(dest_dir, "conversion_info"), exist_ok=True)
        version_number = len(os.listdir(path.join(dest_dir, "conversion_info")))
        conversion_dir = path.join(dest_dir, "conversion_info", f"v{version_number}")
        os.makedirs(conversion_dir)
        cprint(dest_dir, lvl="debug")

        converters = {
            "T1": [adni_t1.convert_adni_t1],
            "PET_FDG": [adni_fdg.convert_adni_fdg_pet],
            "PET_AMYLOID": [
                adni_pib.convert_adni_pib_pet,
                adni_av45_fbb.convert_adni_av45_fbb_pet,
            ],
            "PET_TAU": [adni_tau.convert_adni_tau_pet],
            "DWI": [adni_dwi.convert_adni_dwi],
            "FLAIR": [adni_flair.convert_adni_flair],
            "fMRI": [adni_fmri.convert_adni_fmri],
        }

        for modality in modalities:
            if modality not in converters:
                raise Exception(f"{modality} is not a valid input modality")
            for converter in converters[modality]:
                converter(
                    source_dir,
                    clinical_dir,
                    dest_dir,
                    conversion_dir,
                    subjs_list,
                    force_new_extraction,
                )



# Use hash instead of parameters for iterables folder names
from nipype import config

import clinica.pipelines.engine as cpe

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiConnectome(cpe.Pipeline):
    """Connectome-based processing of corrected DWI datasets.

    Returns:
        A clinica pipeline object containing the DwiConnectome pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        if self.parameters["n_tracks"] < 0:
            raise ValueError(
                f"The n_tracks parameter ({self.parameters['n_tracks']}) should be positive."
            )

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return [
            "t1_brain_file",
            "wm_mask_file",
            "dwi_file",
            "dwi_brainmask_file",
            "grad_fsl",
            "atlas_files",
        ]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return ["response", "fod", "tracts", "nodes", "connectomes"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        import re

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.utils.input_files as input_files
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        # Read CAPS files
        list_caps_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.caps_directory,
            [
                # Inputs from t1-freesurfer pipeline
                input_files.T1_FS_WM,  # list_caps_files[0]
                input_files.T1_FS_DESIKAN,  # list_caps_files[1]
                input_files.T1_FS_DESTRIEUX,  # list_caps_files[2]
                input_files.T1_FS_BRAIN,  # list_caps_files[3]
                # Inputs from dwi-preprocessing pipeline
                input_files.DWI_PREPROC_NII,  # list_caps_files[4]
                input_files.DWI_PREPROC_BRAINMASK,  # list_caps_files[5]
                input_files.DWI_PREPROC_BVEC,  # list_caps_files[6]
                input_files.DWI_PREPROC_BVAL,  # list_caps_files[7]
            ],
            raise_exception=True,
        )

        # Check space of DWI dataset
        dwi_file_spaces = [
            re.search(".*_space-(.*)_preproc.nii.*", file, re.IGNORECASE).group(1)
            for file in list_caps_files[4]
        ]

        # Return an error if all the DWI files are not in the same space
        if any(a != dwi_file_spaces[0] for a in dwi_file_spaces):
            raise ClinicaCAPSError(
                "Preprocessed DWI files are not all in the same space. "
                "Please process them separately using the appropriate subjects/sessions `.tsv` file (-tsv option)."
            )
        list_atlas_files = [
            [aparc_aseg, aparc_aseg_a2009]
            for aparc_aseg, aparc_aseg_a2009 in zip(
                list_caps_files[1], list_caps_files[2]
            )
        ]

        list_grad_fsl = [
            (bvec, bval) for bvec, bval in zip(list_caps_files[6], list_caps_files[7])
        ]

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_participants_sessions(
            self.subjects, self.sessions, folder_participants_tsv
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                "Computational time will depend of the number of volumes in your DWI dataset and "
                "the number of streamlines you selected."
            )

        if dwi_file_spaces[0] == "b0":
            self.parameters["dwi_space"] = "b0"
            read_node = npe.Node(
                name="ReadingFiles",
                iterables=[
                    ("wm_mask_file", list_caps_files[0]),
                    ("t1_brain_file", list_caps_files[3]),
                    ("dwi_file", list_caps_files[4]),
                    ("dwi_brainmask_file", list_caps_files[5]),
                    ("grad_fsl", list_grad_fsl),
                    ("atlas_files", list_atlas_files),
                ],
                synchronize=True,
                interface=nutil.IdentityInterface(fields=self.get_input_fields()),
            )
            # fmt: off
            self.connect(
                [
                    (read_node, self.input_node, [("t1_brain_file", "t1_brain_file")]),
                    (read_node, self.input_node, [("wm_mask_file", "wm_mask_file")]),
                    (read_node, self.input_node, [("dwi_file", "dwi_file")]),
                    (read_node, self.input_node, [("dwi_brainmask_file", "dwi_brainmask_file")]),
                    (read_node, self.input_node, [("grad_fsl", "grad_fsl")]),
                    (read_node, self.input_node, [("atlas_files", "atlas_files")]),
                ]
            )
            # fmt: on
        elif dwi_file_spaces[0] == "T1w":
            self.parameters["dwi_space"] = "T1w"
            read_node = npe.Node(
                name="ReadingFiles",
                iterables=[
                    ("wm_mask_file", list_caps_files[0]),
                    ("dwi_file", list_caps_files[4]),
                    ("dwi_brainmask_file", list_caps_files[5]),
                    ("grad_fsl", list_grad_fsl),
                    ("atlas_files", list_atlas_files),
                ],
                synchronize=True,
                interface=nutil.IdentityInterface(fields=self.get_input_fields()),
            )
            # fmt: off
            self.connect(
                [
                    (read_node, self.input_node, [("wm_mask_file", "wm_mask_file")]),
                    (read_node, self.input_node, [("dwi_file", "dwi_file")]),
                    (read_node, self.input_node, [("dwi_brainmask_file", "dwi_brainmask_file")]),
                    (read_node, self.input_node, [("grad_fsl", "grad_fsl")]),
                    (read_node, self.input_node, [("atlas_files", "atlas_files")]),
                ]
            )
            # fmt: on
        else:
            raise ClinicaCAPSError(
                "Bad preprocessed DWI space. Please check your CAPS folder."
            )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils

        # Writing CAPS
        # ============
        join_node = npe.JoinNode(
            name="JoinOutputs",
            joinsource="ReadingFiles",
            interface=nutil.IdentityInterface(fields=self.get_output_fields()),
        )

        write_node = npe.MapNode(
            name="WritingCAPS",
            iterfield=["container"]
            + [f"connectome_based_processing.@{o}" for o in self.get_output_fields()],
            interface=nio.DataSink(
                infields=[
                    f"connectome_based_processing.@{o}"
                    for o in self.get_output_fields()
                ]
            ),
        )
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.container = utils.get_containers(self.subjects, self.sessions)
        write_node.inputs.substitutions = [("trait_added", "")]
        write_node.inputs.parameterization = False

        # fmt: off
        self.connect(
            [
                # Writing CAPS
                (self.output_node, join_node, [("response", "response")]),
                (self.output_node, join_node, [("fod", "fod")]),
                (self.output_node, join_node, [("tracts", "tracts")]),
                (self.output_node, join_node, [("nodes", "nodes")]),
                (self.output_node, join_node, [("connectomes", "connectomes")]),
                (join_node, write_node, [("response", "connectome_based_processing.@response")]),
                (join_node, write_node, [("fod", "connectome_based_processing.@fod")]),
                (join_node, write_node, [("tracts", "connectome_based_processing.@tracts")]),
                (join_node, write_node, [("nodes", "connectome_based_processing.@nodes")]),
                (join_node, write_node, [("connectomes", "connectome_based_processing.@connectomes")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.

        Notes:
            - If `FSLOUTPUTTYPE` environment variable is not set, `nipype` takes
            NIFTI by default.

        Todo:
            - [x] Detect space automatically.
            - [ ] Allow for custom parcellations (See TODOs in utils).

        """
        import nipype.interfaces.freesurfer as fs
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as niu
        import nipype.pipeline.engine as npe
        from nipype.interfaces.mrtrix3 import (
            ConstrainedSphericalDeconvolution,
            Tractography,
        )
        from nipype.interfaces.mrtrix.preprocess import MRTransform

        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils
        from clinica.utils.exceptions import ClinicaCAPSError
        from clinica.utils.mri_registration import (
            convert_flirt_transformation_to_mrtrix_transformation,
        )

        # Nodes
        # =====
        # B0 Extraction (only if space=b0)
        # -------------
        split_node = npe.Node(name="Reg-0-DWI-B0Extraction", interface=fsl.Split())
        split_node.inputs.output_type = "NIFTI_GZ"
        split_node.inputs.dimension = "t"
        select_node = npe.Node(name="Reg-0-DWI-B0Selection", interface=niu.Select())
        select_node.inputs.index = 0

        # B0 Brain Extraction (only if space=b0)
        # -------------------
        mask_node = npe.Node(name="Reg-0-DWI-BrainMasking", interface=fsl.ApplyMask())
        mask_node.inputs.output_type = "NIFTI_GZ"

        # T1-to-B0 Registration (only if space=b0)
        # ---------------------
        t12b0_reg_node = npe.Node(
            name="Reg-1-T12B0Registration",
            interface=fsl.FLIRT(
                dof=6,
                interp="spline",
                cost="normmi",
                cost_func="normmi",
            ),
        )
        t12b0_reg_node.inputs.output_type = "NIFTI_GZ"

        # MGZ File Conversion (only if space=b0)
        # -------------------
        t1_brain_conv_node = npe.Node(
            name="Reg-0-T1-T1BrainConvertion", interface=fs.MRIConvert()
        )
        wm_mask_conv_node = npe.Node(
            name="Reg-0-T1-WMMaskConvertion", interface=fs.MRIConvert()
        )

        # WM Transformation (only if space=b0)
        # -----------------
        wm_transform_node = npe.Node(
            name="Reg-2-WMTransformation", interface=fsl.ApplyXFM()
        )
        wm_transform_node.inputs.apply_xfm = True

        # Nodes Generation
        # ----------------
        label_convert_node = npe.MapNode(
            name="0-LabelsConversion",
            iterfield=["in_file", "in_config", "in_lut", "out_file"],
            interface=mrtrix3.LabelConvert(),
        )
        label_convert_node.inputs.in_config = utils.get_conversion_luts()
        label_convert_node.inputs.in_lut = utils.get_luts()

        # FSL flirt matrix to MRtrix matrix Conversion (only if space=b0)
        # --------------------------------------------
        fsl2mrtrix_conv_node = npe.Node(
            name="Reg-2-FSL2MrtrixConversion",
            interface=niu.Function(
                input_names=[
                    "in_source_image",
                    "in_reference_image",
                    "in_flirt_matrix",
                    "name_output_matrix",
                ],
                output_names=["out_mrtrix_matrix"],
                function=convert_flirt_transformation_to_mrtrix_transformation,
            ),
        )

        # Parc. Transformation (only if space=b0)
        # --------------------
        parc_transform_node = npe.MapNode(
            name="Reg-2-ParcTransformation",
            iterfield=["in_files", "out_filename"],
            interface=MRTransform(),
        )

        # Response Estimation
        # -------------------
        resp_estim_node = npe.Node(
            name="1a-ResponseEstimation", interface=mrtrix3.ResponseSD()
        )
        resp_estim_node.inputs.algorithm = "tournier"

        # FOD Estimation
        # --------------
        fod_estim_node = npe.Node(
            name="1b-FODEstimation",
            interface=ConstrainedSphericalDeconvolution(),
        )
        fod_estim_node.inputs.algorithm = "csd"

        # Tracts Generation
        # -----------------
        tck_gen_node = npe.Node(name="2-TractsGeneration", interface=Tractography())
        tck_gen_node.inputs.select = self.parameters["n_tracks"]
        tck_gen_node.inputs.algorithm = "iFOD2"

        # Connectome Generation
        # ---------------------
        # only the parcellation and output filename should be iterable, the tck
        # file stays the same.
        conn_gen_node = npe.MapNode(
            name="3-ConnectomeGeneration",
            iterfield=["in_parc", "out_file"],
            interface=mrtrix3.BuildConnectome(),
        )

        # Print begin message
        # -------------------
        print_begin_message = npe.MapNode(
            interface=niu.Function(
                input_names=["in_bids_or_caps_file"],
                function=utils.print_begin_pipeline,
            ),
            iterfield="in_bids_or_caps_file",
            name="WriteBeginMessage",
        )

        # Print end message
        # -----------------
        print_end_message = npe.MapNode(
            interface=niu.Function(
                input_names=["in_bids_or_caps_file", "final_file"],
                function=utils.print_end_pipeline,
            ),
            iterfield=["in_bids_or_caps_file"],
            name="WriteEndMessage",
        )

        # CAPS File names Generation
        # --------------------------
        caps_filenames_node = npe.Node(
            name="CAPSFilenamesGeneration",
            interface=niu.Function(
                input_names="dwi_file",
                output_names=self.get_output_fields(),
                function=utils.get_caps_filenames,
            ),
        )

        # Connections
        # ===========
        # Computation of the diffusion model, tractography & connectome
        # -------------------------------------------------------------
        # fmt: off
        self.connect(
            [
                (self.input_node, print_begin_message, [("dwi_file", "in_bids_or_caps_file")]),
                (self.input_node, caps_filenames_node, [("dwi_file", "dwi_file")]),
                # Response Estimation
                (self.input_node, resp_estim_node, [("dwi_file", "in_file")]),  # Preproc. DWI
                (self.input_node, resp_estim_node, [("dwi_brainmask_file", "in_mask")]),  # B0 brain mask
                (self.input_node, resp_estim_node, [("grad_fsl", "grad_fsl")]),  # bvecs and bvals
                (caps_filenames_node, resp_estim_node, [("response", "wm_file")]),  # output response filename
                # FOD Estimation
                (self.input_node, fod_estim_node, [("dwi_file", "in_file")]),  # Preproc. DWI
                (resp_estim_node, fod_estim_node, [("wm_file", "wm_txt")]),  # Response (txt file)
                (self.input_node, fod_estim_node, [("dwi_brainmask_file", "mask_file")]),  # B0 brain mask
                (self.input_node, fod_estim_node, [("grad_fsl", "grad_fsl")]),  # T1-to-B0 matrix file
                (caps_filenames_node, fod_estim_node, [("fod", "wm_odf")]),  # output odf filename
                # Tracts Generation
                (fod_estim_node, tck_gen_node, [("wm_odf", "in_file")]),  # ODF file
                (caps_filenames_node, tck_gen_node, [("tracts", "out_file")]),  # output tck filename
                # Label Conversion
                (self.input_node, label_convert_node, [("atlas_files", "in_file")]),  # atlas image files
                (caps_filenames_node, label_convert_node, [("nodes", "out_file")]),  # converted atlas image filenames
                # Connectomes Generation
                (tck_gen_node, conn_gen_node, [("out_file", "in_file")]),
                (caps_filenames_node, conn_gen_node, [("connectomes", "out_file")]),
            ]
        )
        # Registration T1-DWI (only if space=b0)
        # -------------------
        if self.parameters["dwi_space"] == "b0":
            self.connect(
                [
                    # MGZ Files Conversion
                    (self.input_node, t1_brain_conv_node, [("t1_brain_file", "in_file")]),
                    (self.input_node, wm_mask_conv_node, [("wm_mask_file", "in_file")]),
                    # B0 Extraction
                    (self.input_node, split_node, [("dwi_file", "in_file")]),
                    (split_node, select_node, [("out_files", "inlist")]),
                    # Masking
                    (select_node, mask_node, [("out", "in_file")]),  # B0
                    (self.input_node, mask_node, [("dwi_brainmask_file", "mask_file")]),  # Brain mask
                    # T1-to-B0 Registration
                    (t1_brain_conv_node, t12b0_reg_node, [("out_file", "in_file")]),  # Brain
                    (mask_node, t12b0_reg_node, [("out_file", "reference")]),  # B0 brain-masked
                    # WM Transformation
                    (wm_mask_conv_node, wm_transform_node, [("out_file", "in_file")]),  # Brain mask
                    (mask_node, wm_transform_node, [("out_file", "reference")]),  # BO brain-masked
                    (t12b0_reg_node, wm_transform_node, [("out_matrix_file", "in_matrix_file")]),  # T1-to-B0 matrix file
                    # FSL flirt matrix to MRtrix matrix Conversion
                    (t1_brain_conv_node, fsl2mrtrix_conv_node, [("out_file", "in_source_image")]),
                    (mask_node, fsl2mrtrix_conv_node, [("out_file", "in_reference_image")]),
                    (t12b0_reg_node, fsl2mrtrix_conv_node, [("out_matrix_file", "in_flirt_matrix")]),
                    # Apply registration without resampling on parcellations
                    (label_convert_node, parc_transform_node, [("out_file", "in_files")]),
                    (fsl2mrtrix_conv_node, parc_transform_node, [("out_mrtrix_matrix", "linear_transform")]),
                    (caps_filenames_node, parc_transform_node, [("nodes", "out_filename")]),
                ]
            )
        # Special care for Parcellation & WM mask
        # ---------------------------------------
        if self.parameters["dwi_space"] == "b0":
            self.connect(
                [
                    (wm_transform_node, tck_gen_node, [("out_file", "seed_image")]),
                    (parc_transform_node, conn_gen_node, [("out_file", "in_parc")]),
                    (parc_transform_node, self.output_node, [("out_file", "nodes")]),
                ]
            )
        elif self.parameters["dwi_space"] == "T1w":
            self.connect(
                [
                    (self.input_node, tck_gen_node, [("wm_mask_file", "seed_image")]),
                    (label_convert_node, conn_gen_node, [("out_file", "in_parc")]),
                    (label_convert_node, self.output_node, [("out_file", "nodes")]),
                ]
            )
        else:
            raise ClinicaCAPSError(
                "Bad preprocessed DWI space. Please check your CAPS folder."
            )
        # Outputs
        # -------
        self.connect(
            [
                (resp_estim_node, self.output_node, [("wm_file", "response")]),
                (fod_estim_node, self.output_node, [("wm_odf", "fod")]),
                (tck_gen_node, self.output_node, [("out_file", "tracts")]),
                (conn_gen_node, self.output_node, [("out_file", "connectomes")]),
                (self.input_node, print_end_message, [("dwi_file", "in_bids_or_caps_file")]),
                (conn_gen_node, print_end_message, [("out_file", "final_file")]),
            ]
        )
        # fmt: off



"""Convert the NIFD dataset into BIDS."""

from os import PathLike
from typing import List


def convert_images(
    path_to_dataset: PathLike,
    bids_dir: PathLike,
    path_to_clinical: PathLike,
) -> List[PathLike]:
    """Convert the entire dataset in BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """

    import clinica.iotools.bids_utils as bids

    from .nifd_utils import (
        dataset_to_bids,
        read_clinical_data,
        read_imaging_data,
        write_bids,
    )

    clinical_data = read_clinical_data(path_to_clinical)
    imaging_data = read_imaging_data(path_to_dataset)

    participants, sessions, scans = dataset_to_bids(
        imaging_data=imaging_data, clinical_data=clinical_data
    )

    written = write_bids(
        to=bids_dir,
        participants=participants,
        sessions=sessions,
        scans=scans,
    )
    readme_data = {
        "link": "https://ida.loni.usc.edu/home/projectPage.jsp?project=NIFD&page=HOME&subPage=OVERVIEW_PR#",
        "desc": (
            "NIFD is the nickname for the frontotemporal lobar degeneration neuroimaging initiative "
            "(FTLDNI, AG032306), which was funded by the NIA and NINDS to characterize longitudinal clinical and "
            "imaging changes in FTLD.The imaging and clinical methods are the same for NIFD and for the 4-Repeat "
            "Tauopathy Neuroimaging Initiative (4RTNI), which is also available for download from LONI. Controls for "
            "NIFD are the same controls as those collected for 4RTNI."
        ),
    }
    bids.write_modality_agnostic_files(
        study_name="NIFD", readme_data=readme_data, bids_dir=bids_dir
    )
    return written



import os

from nipype.interfaces.base import File, TraitedSpec, traits
from nipype.interfaces.spm.base import SPMCommand, SPMCommandInputSpec, scans_for_fnames
from nipype.utils.filemanip import split_filename

from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

if spm_standalone_is_available():
    use_spm_standalone()


class DARTELExistingTemplateInputSpec(SPMCommandInputSpec):
    image_files = traits.List(
        traits.List(File(exists=True)),
        desc="A list of files to be segmented",
        field="warp1.images",
        copyfile=False,
        mandatory=True,
    )
    regularization_form = traits.Enum(
        "Linear",
        "Membrane",
        "Bending",
        field="warp1.settings.rform",
        desc="Form of regularization energy term",
    )
    iteration_parameters = traits.List(
        traits.Tuple(
            traits.Range(1, 10),
            traits.Tuple(traits.Float, traits.Float, traits.Float),
            traits.Range(0, 9),
            traits.File(exists=True),
        ),
        minlen=3,
        maxlen=12,
        mandatory=True,
        field="warp1.settings.param",
        desc="""List of tuples for each iteration
                                       - Inner iterations
                                       - Regularization parameters
                                       - Time points for deformation model
                                       - DARTEL template
                                       """,
    )
    optimization_parameters = traits.Tuple(
        traits.Float,
        traits.Range(1, 8),
        traits.Range(1, 8),
        field="warp1.settings.optim",
        desc="""Optimization settings a tuple
                                           - LM regularization
                                           - cycles of multigrid solver
                                           - relaxation iterations
                                           """,
    )


class DARTELExistingTemplateOutputSpec(TraitedSpec):
    dartel_flow_fields = traits.List(File(exists=True), desc="DARTEL flow fields")


class DARTELExistingTemplate(SPMCommand):
    """Use SPM DARTEL to create a template and flow fields
    http://www.fil.ion.ucl.ac.uk/spm/doc/manual.pdf#page=185
    Examples
    --------
    >>> import nipype.interfaces.spm as spm
    >>> dartel = spm.DARTEL()
    >>> dartel.inputs.image_files = [['rc1s1.nii','rc1s2.nii'],['rc2s1.nii', 'rc2s2.nii']]
    >>> dartel.run() # doctest: +SKIP
    """

    input_spec = DARTELExistingTemplateInputSpec
    output_spec = DARTELExistingTemplateOutputSpec
    _jobtype = "tools"
    _jobname = "dartel"

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm"""
        import numpy as np

        if opt in ["image_files"]:
            return scans_for_fnames(val, keep4d=True, separate_sessions=True)
        elif opt == "regularization_form":
            mapper = {"Linear": 0, "Membrane": 1, "Bending": 2}
            return mapper[val]
        elif opt == "iteration_parameters":
            params = []
            for param in val:
                new_param = {}
                new_param["its"] = param[0]
                new_param["rparam"] = list(param[1])
                new_param["K"] = param[2]
                new_param["template"] = np.array([param[3]], dtype=object)
                params.append(new_param)
            return params
        elif opt == "optimization_parameters":
            new_param = {}
            new_param["lmreg"] = val[0]
            new_param["cyc"] = val[1]
            new_param["its"] = val[2]
            return [new_param]
        else:
            return super(DARTELExistingTemplate, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["dartel_flow_fields"] = []
        for filename in self.inputs.image_files[0]:
            pth, base, ext = split_filename(filename)
            outputs["dartel_flow_fields"].append(
                os.path.realpath("u_%s%s" % (base, ext))
            )
        return outputs


def prepare_dartel_input_images(dartel_input_images):
    return [[[tissue] for tissue in subject] for subject in zip(*dartel_input_images)]


def create_iteration_parameters(dartel_templates, iteration_parameters):

    if len(dartel_templates) != 6:
        raise ValueError(
            "Wrong number of templates. 6 templates were expected, one for each DARTEL iteration."
        )

    if not iteration_parameters:
        iter1 = (3, (4, 2, 1e-06), 0, dartel_templates[0])
        iter2 = (3, (2, 1, 1e-06), 0, dartel_templates[1])
        iter3 = (3, (1, 0.5, 1e-06), 1, dartel_templates[2])
        iter4 = (3, (0.5, 0.25, 1e-06), 2, dartel_templates[3])
        iter5 = (3, (0.25, 0.125, 1e-06), 4, dartel_templates[4])
        iter6 = (3, (0.25, 0.125, 1e-06), 6, dartel_templates[5])

        return [iter1, iter2, iter3, iter4, iter5, iter6]

    elif len(iteration_parameters) != 6:
        raise ValueError(
            "Wrong number of iteration parameters. 6 iterations were expected."
        )
    else:
        new_iteration_parameters = []
        for i in range(6):
            new_iteration_parameters.append(
                (
                    iteration_parameters[i][0],
                    iteration_parameters[i][1],
                    iteration_parameters[i][2],
                    dartel_templates[i],
                )
            )
        return new_iteration_parameters



from pathlib import Path

import click

import clinica.iotools.bids_utils as bids


@click.command(name="habs-to-bids")
@click.argument(
    "sourcedata",
    type=click.Path(exists=True, resolve_path=True, path_type=Path),
)
@click.argument(
    "rawdata",
    type=click.Path(writable=True, resolve_path=True, path_type=Path),
)
def cli(sourcedata: str, rawdata: str) -> None:
    """HABS to BIDS converter."""
    from pandas import concat

    from .habs_to_bids import (
        find_clinical_data,
        find_imaging_data,
        parse_imaging_data,
        read_clinical_data,
        write_bids,
    )

    clinical_data = {
        k: read_clinical_data(sourcedata / p, c)
        for k, p, c in find_clinical_data(sourcedata)
    }

    imaging_data = concat(
        [parse_imaging_data(x) for x in find_imaging_data(sourcedata)]
    )

    write_bids(
        sourcedata=sourcedata,
        rawdata=rawdata,
        imaging_data=imaging_data,
        clinical_data=clinical_data,
    )


if __name__ == "__main__":
    cli()



from typing import IO

from attrs import define


@define
class BIDSReadme:
    """Model representing the content for a BIDS Readme."""

    name: str
    link: str
    description: str

    def write(self, to: IO[str]):
        from importlib.metadata import version

        to.write(
            f"This BIDS directory was generated with Clinica v{version('clinica')}.\n"
            f"More information on https://www.clinica.run\n"
            f"\n"
            f"Study: {self.name}\n"
            f"\n"
            f"{self.description}\n\n"
            f"Find more about it and about the data user agreement: {self.link}"
        )



from nipype import config

import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiPreprocessingUsingT1(cpe.Pipeline):
    """DWI Preprocessing using T1 image for susceptibility distortion step.

    Ideas for improvement:
        - Replace prepare_reference_b0 function by a first run of FSL eddy
        - Replace B0-T1w registration by FA-T1w registration
        - Replace os.system(cmd) by call of Nipype interface
        - Use promising sdcflows workflows and/or dMRIprep

    Warnings:
        - Do not use this pipeline if you have fieldmap data in your dataset.

    Note:
        Some reading regarding the reproducibility of FSL eddy command:
        https://www.jiscmail.ac.uk/cgi-bin/webadmin?A2=fsl;1ccf038f.1608

    Returns:
        A clinica pipeline object containing the DWIPreprocessingUsingT1 pipeline.
    """

    @staticmethod
    def get_processed_images(caps_directory, subjects, sessions):
        import os

        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import DWI_PREPROC_NII
        from clinica.utils.inputs import clinica_file_reader

        image_ids = []
        if os.path.isdir(caps_directory):
            preproc_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, DWI_PREPROC_NII, False
            )
            image_ids = extract_image_ids(preproc_files)
        return image_ids

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.stream import cprint

        self.parameters.setdefault("low_bval", 5)
        low_bval = self.parameters["low_bval"]
        if low_bval < 0:
            raise ValueError(
                f"The low_bval is negative ({low_bval}): it should be zero or close to zero."
            )
        if self.parameters["low_bval"] > 100:
            cprint(
                f"The low_bval parameter is {low_bval}: it should be close to zero.",
                lvl="warning",
            )

        self.parameters.setdefault("use_cuda", False)
        self.parameters.setdefault("initrand", False)
        self.parameters.setdefault("delete_cache", False)

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        from clinica.utils.check_dependency import is_binary_present
        from clinica.utils.exceptions import ClinicaMissingDependencyError

        if self.parameters["use_cuda"]:
            if not is_binary_present("eddy_cuda"):
                raise ClinicaMissingDependencyError(
                    "[Error] FSL eddy with CUDA was set but Clinica could not find eddy_cuda in your PATH environment. "
                    "Check that  https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/eddy/UsersGuide#The_eddy_executables is correctly set."
                )

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            List[str]: The list of inputs for the DwiPreprocessingUsingT1 pipeline namely:
                * t1w: Path of the T1w image in BIDS format
                * dwi: Path of the diffusion weighted image in BIDS format
                * dwi_json: Path of the DWI JSON file in BIDS format and containing
                    TotalReadoutTime and PhaseEncodingDirection metadata (see BIDS specifications)
                * bvec: Path of the bvec file in BIDS format
                * bval: Path of the bval file in BIDS format
        """
        return ["t1w", "dwi", "dwi_json", "bvec", "bval"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            List[str]: The list of outputs generated by the DwiPreprocessingUsingT1 pipeline namely:
                * preproc_dwi: Path of the preprocessed DWI
                * preproc_bvec: Path of the preprocessed bvec
                * preproc_bval: Path of the preprocessed bval
                * b0_mask: Path of the b0 brainmask
        """
        return ["preproc_dwi", "preproc_bvec", "preproc_bval", "b0_mask"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.input_files import (
            DWI_BVAL,
            DWI_BVEC,
            DWI_JSON,
            DWI_NII,
            T1W_NII,
        )
        from clinica.utils.inputs import clinica_list_of_files_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        list_bids_files = clinica_list_of_files_reader(
            self.subjects,
            self.sessions,
            self.bids_directory,
            [T1W_NII, DWI_JSON, DWI_NII, DWI_BVEC, DWI_BVAL],
            raise_exception=True,
        )

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_participants_sessions(
            self.subjects, self.sessions, folder_participants_tsv
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                f"List available in {os.path.join(folder_participants_tsv, 'participants.tsv')}"
            )
            cprint(
                "Computational time will depend of the number of volumes in your DWI dataset and the use of CUDA."
            )

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("t1w", list_bids_files[0]),
                ("dwi_json", list_bids_files[1]),
                ("dwi", list_bids_files[2]),
                ("bvec", list_bids_files[3]),
                ("bval", list_bids_files[4]),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        # fmt: off
        self.connect(
            [
                (read_node, self.input_node, [("t1w", "t1w"),
                                              ("dwi", "dwi"),
                                              ("dwi_json", "dwi_json"),
                                              ("bvec", "bvec"),
                                              ("bval", "bval")]),
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.nipype import container_from_filename, fix_join

        from .dwi_preprocessing_using_t1_utils import rename_into_caps

        # Find container path from DWI filename
        # =====================================
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="container_path",
        )

        rename_into_caps = npe.Node(
            nutil.Function(
                input_names=[
                    "in_bids_dwi",
                    "fname_dwi",
                    "fname_bval",
                    "fname_bvec",
                    "fname_brainmask",
                ],
                output_names=[
                    "out_caps_dwi",
                    "out_caps_bval",
                    "out_caps_bvec",
                    "out_caps_brainmask",
                ],
                function=rename_into_caps,
            ),
            name="rename_into_caps",
        )

        # Writing results into CAPS
        # =========================
        write_results = npe.Node(name="write_results", interface=nio.DataSink())
        write_results.inputs.base_directory = self.caps_directory
        write_results.inputs.parameterization = False

        # fmt: off
        self.connect(
            [
                (self.input_node, container_path, [("dwi", "bids_or_caps_filename")]),
                (self.input_node, rename_into_caps, [("dwi", "in_bids_dwi")]),
                (self.output_node, rename_into_caps, [("preproc_dwi", "fname_dwi"),
                                                      ("preproc_bval", "fname_bval"),
                                                      ("preproc_bvec", "fname_bvec"),
                                                      ("b0_mask", "fname_brainmask")]),
                (container_path, write_results, [(("container", fix_join, "dwi"), "container")]),
                (rename_into_caps, write_results, [("out_caps_dwi", "preprocessing.@preproc_dwi"),
                                                   ("out_caps_bval", "preprocessing.@preproc_bval"),
                                                   ("out_caps_bvec", "preprocessing.@preproc_bvec"),
                                                   ("out_caps_brainmask", "preprocessing.@b0_mask")])
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.dwi import compute_average_b0

        from .dwi_preprocessing_using_t1_utils import (
            init_input_node,
            prepare_reference_b0,
            print_end_pipeline,
        )
        from .dwi_preprocessing_using_t1_workflows import (
            eddy_fsl_pipeline,
            epi_pipeline,
        )

        # Nodes creation
        # ==============
        # Initialize input parameters and print begin message
        init_node = npe.Node(
            interface=nutil.Function(
                input_names=self.get_input_fields(),
                output_names=[
                    "image_id",
                    "t1w",
                    "dwi",
                    "bvec",
                    "bval",
                    "total_readout_time",
                    "phase_encoding_direction",
                ],
                function=init_input_node,
            ),
            name="0-InitNode",
        )

        # Prepare b0 image for further corrections
        prepare_b0 = npe.Node(
            name="PrepareB0",
            interface=nutil.Function(
                input_names=[
                    "in_dwi",
                    "in_bval",
                    "in_bvec",
                    "low_bval",
                    "working_directory",
                ],
                output_names=[
                    "out_reference_b0",
                    "out_b0_dwi_merge",
                    "out_updated_bval",
                    "out_updated_bvec",
                ],
                function=prepare_reference_b0,
            ),
        )
        prepare_b0.inputs.low_bval = self.parameters["low_bval"]
        prepare_b0.inputs.working_directory = self.base_dir
        # Mask b0 for computations purposes
        mask_b0_pre = npe.Node(
            fsl.BET(frac=0.3, mask=True, robust=True), name="PreMaskB0"
        )
        # Head-motion correction + Eddy-currents correction
        eddy_fsl = eddy_fsl_pipeline(
            low_bval=self.parameters["low_bval"],
            use_cuda=self.parameters["use_cuda"],
            initrand=self.parameters["initrand"],
        )
        # Susceptibility distortion correction using T1w image
        sdc = epi_pipeline(
            self.base_dir,
            self.parameters["delete_cache"],
            name="SusceptibilityDistortionCorrection",
        )

        # Remove bias correction from (Jeurissen et al., 2014)
        bias = npe.Node(
            mrtrix3.DWIBiasCorrect(
                use_ants=True,
                out_file="Jacobian_image_maths_thresh_merged_biascorr.nii.gz",
            ),
            name="RemoveBias",
        )
        # Compute b0 mask on corrected avg b0
        compute_avg_b0 = npe.Node(
            nutil.Function(
                input_names=["in_dwi", "in_bval"],
                output_names=["out_b0_average"],
                function=compute_average_b0,
            ),
            name="ComputeB0Average",
        )
        compute_avg_b0.inputs.low_bval = self.parameters["low_bval"]

        # Compute brain mask from reference b0
        mask_avg_b0 = npe.Node(fsl.BET(mask=True, robust=True), name="MaskB0")

        # Print end message
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["image_id", "final_file"], function=print_end_pipeline
            ),
            name="WriteEndMessage",
        )

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                # Initialize input parameters and print begin message
                (self.input_node, init_node, [("t1w", "t1w"),
                                              ("dwi", "dwi"),
                                              ("bvec", "bvec"),
                                              ("bval", "bval"),
                                              ("dwi_json", "dwi_json")]),
                # Preliminary step (possible computation of a mean b0):
                (init_node, prepare_b0, [("dwi", "in_dwi"),
                                         ("bval", "in_bval"),
                                         ("bvec", "in_bvec")]),
                # Mask b0 before corrections
                (prepare_b0, mask_b0_pre, [("out_reference_b0", "in_file")]),
                # Head-motion correction + eddy current correction
                (init_node, eddy_fsl, [("total_readout_time", "inputnode.total_readout_time"),
                                       ("phase_encoding_direction", "inputnode.phase_encoding_direction")]),
                (prepare_b0, eddy_fsl, [("out_b0_dwi_merge", "inputnode.in_file"),
                                        ("out_updated_bval", "inputnode.in_bval"),
                                        ("out_updated_bvec", "inputnode.in_bvec"),
                                        ("out_reference_b0", "inputnode.ref_b0")]),
                (mask_b0_pre, eddy_fsl, [("mask_file", "inputnode.in_mask")]),
                # Magnetic susceptibility correction
                (init_node, sdc, [("t1w", "inputnode.T1")]),
                (eddy_fsl, sdc, [("outputnode.out_corrected", "inputnode.DWI")]),
                (eddy_fsl, sdc, [("outputnode.out_rotated_bvecs", "inputnode.bvec")]),
                # Bias correction
                (prepare_b0, bias, [("out_updated_bval", "in_bval")]),
                (sdc, bias, [("outputnode.DWIs_epicorrected", "in_file"),
                             ("outputnode.out_bvec", "in_bvec")]),
                # Compute average b0 on corrected dataset (for brain mask extraction)
                (prepare_b0, compute_avg_b0, [("out_updated_bval", "in_bval")]),
                (bias, compute_avg_b0, [("out_file", "in_dwi")]),
                # Compute b0 mask on corrected avg b0
                (compute_avg_b0, mask_avg_b0, [("out_b0_average", "in_file")]),
                # Print end message
                (init_node, print_end_message, [("image_id", "image_id")]),
                (mask_avg_b0, print_end_message, [("mask_file", "final_file")]),
                # Output node
                (bias, self.output_node, [("out_file", "preproc_dwi")]),
                (sdc, self.output_node, [("outputnode.out_bvec", "preproc_bvec")]),
                (prepare_b0, self.output_node, [("out_updated_bval", "preproc_bval")]),
                (mask_avg_b0, self.output_node, [("mask_file", "b0_mask")]),
            ]
        )
        # fmt: on



def statistics_on_atlases(in_registered_map, name_map, prefix_file=None):
    """Computes a list of statistics files for each atlas.

    Args:
        in_registered_map (str): Map already registered on atlases.
        name_map (str): Name of the registered map in CAPS format.
        prefix_file (Opt[str]):
            <prefix_file>_space-<atlas_name>_map-<name_map>_statistics.tsv

    Returns:
        List of paths leading to the statistics TSV files.
    """
    import os

    from nipype.utils.filemanip import split_filename

    from clinica.utils.atlas import (
        AtlasAbstract,
        JHUDTI811mm,
        JHUTracts01mm,
        JHUTracts251mm,
    )
    from clinica.utils.statistics import statistics_on_atlas

    in_atlas_list = [JHUDTI811mm(), JHUTracts01mm(), JHUTracts251mm()]

    atlas_statistics_list = []
    for atlas in in_atlas_list:
        if not isinstance(atlas, AtlasAbstract):
            raise TypeError("Atlas element must be an AtlasAbstract type")

        if prefix_file:
            filename = (
                f"{prefix_file}_space-{atlas.get_name_atlas()}"
                f"_res-{atlas.get_spatial_resolution()}_map-{name_map}_statistics.tsv"
            )
        else:
            _, base, _ = split_filename(in_registered_map)
            filename = (
                f"{base}_space-{atlas.get_name_atlas()}"
                f"_res-{atlas.get_spatial_resolution()}_map-{name_map}_statistics.tsv"
            )

        out_atlas_statistics = os.path.abspath(os.path.join(os.getcwd(), filename))

        statistics_on_atlas(in_registered_map, atlas, out_atlas_statistics)
        atlas_statistics_list.append(out_atlas_statistics)

    return atlas_statistics_list


def extract_bids_identifier_from_caps_filename(caps_dwi_filename: str) -> str:
    """Extract BIDS identifier from CAPS filename."""
    import re

    m = re.search(r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi", caps_dwi_filename)

    if not m:
        raise ValueError(
            f"Input filename {caps_dwi_filename} is not in a CAPS compliant format."
        )
    bids_identifier = m.group(0)

    return bids_identifier


def get_caps_filenames(caps_dwi_filename: str):
    """Prepare some filenames with CAPS naming convention."""
    import re

    m = re.search(
        r"(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+).*_dwi_space-[a-zA-Z0-9]+",
        caps_dwi_filename,
    )
    if not m:
        raise ValueError(
            f"Input filename {caps_dwi_filename} is not in a CAPS compliant format."
        )

    caps_prefix = m.group(0)
    bids_source = f"{m.group(1)}_{m.group(2)}_dwi"

    out_dti = f"{caps_prefix}_model-DTI_diffmodel.nii.gz"
    out_fa = f"{caps_prefix}_FA.nii.gz"
    out_md = f"{caps_prefix}_MD.nii.gz"
    out_ad = f"{caps_prefix}_AD.nii.gz"
    out_rd = f"{caps_prefix}_RD.nii.gz"
    out_evec = f"{caps_prefix}_DECFA.nii.gz"

    return bids_source, out_dti, out_fa, out_md, out_ad, out_rd, out_evec


def rename_into_caps(
    in_caps_dwi,
    in_norm_fa,
    in_norm_md,
    in_norm_ad,
    in_norm_rd,
    in_b_spline_transform,
    in_affine_matrix,
):
    """Rename different outputs of the pipelines into CAPS format.

    Returns:
        The different outputs with CAPS naming convention
    """
    from nipype.interfaces.utility import Rename

    from clinica.pipelines.dwi_dti.dwi_dti_utils import (
        extract_bids_identifier_from_caps_filename,
    )

    bids_identifier = extract_bids_identifier_from_caps_filename(in_caps_dwi)

    # CAPS normalized FA
    rename_fa = Rename()
    rename_fa.inputs.in_file = in_norm_fa
    rename_fa.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_FA.nii.gz"
    )
    out_caps_fa = rename_fa.run()
    # CAPS normalized MD
    rename_md = Rename()
    rename_md.inputs.in_file = in_norm_md
    rename_md.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_MD.nii.gz"
    )
    out_caps_md = rename_md.run()
    # CAPS normalized AD
    rename_ad = Rename()
    rename_ad.inputs.in_file = in_norm_ad
    rename_ad.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_AD.nii.gz"
    )
    out_caps_ad = rename_ad.run()
    # CAPS normalized RD
    rename_rd = Rename()
    rename_rd.inputs.in_file = in_norm_rd
    rename_rd.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_RD.nii.gz"
    )
    out_caps_rd = rename_rd.run()
    # CAPS B-spline transform
    rename_b_spline = Rename()
    rename_b_spline.inputs.in_file = in_b_spline_transform
    rename_b_spline.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_deformation.nii.gz"
    )
    out_caps_b_spline_transform = rename_b_spline.run()
    # CAPS Affine matrix
    rename_affine = Rename()
    rename_affine.inputs.in_file = in_affine_matrix
    rename_affine.inputs.format_string = (
        f"{bids_identifier}_space-MNI152Lin_res-1x1x1_affine.mat"
    )
    out_caps_affine_matrix = rename_affine.run()

    return (
        out_caps_fa.outputs.out_file,
        out_caps_md.outputs.out_file,
        out_caps_ad.outputs.out_file,
        out_caps_rd.outputs.out_file,
        out_caps_b_spline_transform.outputs.out_file,
        out_caps_affine_matrix.outputs.out_file,
    )


def print_begin_pipeline(in_bids_or_caps_file):
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    print_begin_image(get_subject_id(in_bids_or_caps_file))


def print_end_pipeline(in_bids_or_caps_file, final_file_1, final_file_2):
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(in_bids_or_caps_file))


def get_ants_transforms(in_affine_transformation, in_bspline_transformation):
    """Combine transformations for antsApplyTransforms interface."""
    return [in_bspline_transformation, in_affine_transformation]



from typing import List, Optional

import click

bids_directory = click.argument(
    "bids_directory", type=click.Path(exists=True, resolve_path=True)
)

current_directory = click.argument(
    "current_directory", type=click.Path(exists=True, resolve_path=True)
)


@click.group("iotools")
def cli() -> None:
    """Tools to handle BIDS/CAPS datasets."""
    pass


@cli.command()
@bids_directory
@click.argument("output_bids_directory", type=click.Path(resolve_path=True))
@click.option(
    "-m",
    "--modality",
    "modalities",
    multiple=True,
    metavar="modalities",
    help="Process selected modalities.",
)
@click.option("--center-all-files", is_flag=True, help="Force processing of all files.")
def center_nifti(
    bids_directory: str,
    output_bids_directory: str,
    modalities: Optional[List[str]] = None,
    center_all_files: bool = False,
) -> None:
    """Center NIfTI files in a BIDS dataset.

    This tool is mainly useful as a preprocessing step of SPM. In some cases, SPM is not able to segment T1 volumes
    because their respective center is not aligned with the origin of the world coordinate system. By default, only
    images detected as problematic are converted from INPUT_BIDS_DIRECTORY to OUTPUT_BIDS_DIRECTORY, whilst the others
    are copied verbatim.
    """
    import sys
    import time
    from os import listdir, makedirs
    from os.path import isfile, join

    from clinica.iotools.utils.data_handling import (
        center_all_nifti,
        write_list_of_files,
    )
    from clinica.utils.stream import cprint

    # check that output_folder does not exist, or is an empty folder
    try:
        makedirs(output_bids_directory)
    except FileExistsError:
        file_list = [
            file for file in listdir(output_bids_directory) if not file.startswith(".")
        ]
        if file_list:
            click.echo(
                "Target BIDS directory is not empty. Existing files may be overwritten."
            )
            if not click.confirm("Do you wish to continue?"):
                click.echo("Clinica will now exit...")
                sys.exit(0)

    cprint("Clinica is now centering the requested images.")

    centered_files = center_all_nifti(
        bids_directory,
        output_bids_directory,
        modalities,
        center_all_files,
    )

    # Write list of created files
    timestamp = time.strftime("%Y%m%d-%H%M%S", time.localtime(time.time()))
    log_file = join(output_bids_directory, "centered_nifti_list_" + timestamp + ".txt")
    # If an error happen while creating the file, the function returns Nan
    if not write_list_of_files(centered_files, log_file):
        cprint("Could not create log file.")

    # Final message
    cprint(
        f"{str(len(centered_files))} NIfTI files/images of BIDS folder:\n"
        f"\t{bids_directory}\n"
        f"for the modalities {modalities} have been centered in output folder:\n"
        f"\t{output_bids_directory}"
    )
    if isfile(log_file):
        cprint(f"The list of centered NIfTI files is available here: {log_file}.")
    cprint(
        "Please note that the rest of the input BIDS folder has also been copied to the output folder."
    )


@cli.command()
@bids_directory
@click.argument("output_directory", type=click.Path(writable=True))
@click.option(
    "-op",
    "--output_prefix",
    default="missing_mods",
    show_default=True,
    help="Prefix for the name of the output files.",
)
def check_missing_modalities(
    bids_directory: str,
    output_directory: str,
    output_prefix: str = "missing_mods",
) -> None:
    """Check missing modalities in a BIDS dataset."""
    from clinica.iotools.utils.data_handling import compute_missing_mods
    from clinica.utils.inputs import check_bids_folder

    check_bids_folder(bids_directory)
    compute_missing_mods(bids_directory, output_directory, output_prefix)


@cli.command()
@bids_directory
@click.argument("caps_directory", type=click.Path(exists=True, resolve_path=True))
@click.argument("output_file", type=click.Path(resolve_path=True))
def check_missing_processing(
    bids_directory: str,
    caps_directory: str,
    output_file: str,
) -> None:
    """Check missing processing in a CAPS dataset."""
    from clinica.iotools.utils.data_handling import compute_missing_processing
    from clinica.utils.inputs import check_caps_folder

    check_caps_folder(caps_directory)
    compute_missing_processing(bids_directory, caps_directory, output_file)


@cli.command()
@current_directory
@click.argument("output_tsv", type=click.Path(resolve_path=True))
def create_subjects_visits(current_directory: str, output_tsv: str) -> None:
    """Export participants with their sessions."""
    from os import makedirs
    from os.path import basename, dirname

    from clinica.iotools.utils.data_handling import create_subs_sess_list
    from clinica.utils.inputs import determine_caps_or_bids
    from clinica.utils.stream import cprint

    is_bids = determine_caps_or_bids(current_directory)
    output_directory = dirname(output_tsv)
    makedirs(output_directory, exist_ok=True)
    create_subs_sess_list(
        current_directory, output_directory, basename(output_tsv), is_bids_dir=is_bids
    )
    cprint(f"The TSV file was saved to {output_tsv}.")


@cli.command()
@bids_directory
@click.argument("output_tsv", type=click.Path(resolve_path=True))
@click.option(
    "-caps",
    "--caps_directory",
    type=click.Path(exists=True),
    help="Path to a CAPS dataset.",
)
@click.option(
    "-p",
    "--pipeline",
    "pipelines",
    multiple=True,
    type=click.Choice(
        [
            "t1-freesurfer",
            "t1-volume",
            "pet-volume",
            "t1-freesurfer-longitudinal",
            "dwi-dti",
        ]
    ),
    help="Pipeline to merge to the ouput TSV file. All pipelines are merged by default.",
)
@click.option(
    "-vas",
    "--volume_atlas_selection",
    multiple=True,
    help="Atlas to merge for t1- and pet-volume. All atlases are merged by default.",
)
@click.option(
    "-fas",
    "--freesurfer_atlas_selection",
    multiple=True,
    help="Atlas to merge for t1-freesurfer. All atlases are merged by default",
)
@click.option(
    "-pvc",
    "--pvc_restriction",
    type=click.IntRange(0, 1),
    help=(
        "Restriction on the label [_pvc-rbv] for pet-volume. Default: Merge all atlases, "
        "0: Merge atlases without the label only, 1: Merge atlases with the label only."
    ),
)
@click.option(
    "-pts",
    "--pet_tracers_selection",
    multiple=True,
    help="PET tracer to merge. All PET tracers are merged by default.",
)
@click.option(
    "-group",
    "--group_selection",
    multiple=True,
    help="Group to merge. All groups are merged by default.",
)
@click.option(
    "-tsv",
    "--subjects_sessions_tsv",
    type=click.Path(exists=True, resolve_path=True),
    help="List of subjects and their sessions in TSV format.",
)
@click.option(
    "--ignore_scan_files",
    is_flag=True,
    help="Ignore scan files. This may accelerate the procedure.",
)
@click.option(
    "--ignore_session_scan_files",
    is_flag=True,
    help="Ignore session files. This may accelerate the procedure.",
)
def merge_tsv(
    bids_directory: str,
    output_tsv: str,
    caps_directory: Optional[str] = None,
    pipelines: Optional[List[str]] = None,
    volume_atlas_selection: Optional[List[str]] = None,
    freesurfer_atlas_selection: Optional[List[str]] = None,
    pvc_restriction: Optional[int] = None,
    pet_tracers_selection: Optional[List[str]] = None,
    group_selection: Optional[List[str]] = None,
    subjects_sessions_tsv: Optional[str] = None,
    ignore_scan_files: bool = False,
    ignore_session_scan_files: bool = False,
) -> None:
    """Merge clinical data into a single TSV file."""

    from clinica.iotools.utils.data_handling import create_merge_file
    from clinica.utils.inputs import check_bids_folder

    check_bids_folder(bids_directory)

    create_merge_file(
        bids_directory,
        output_tsv,
        caps_dir=caps_directory,
        pipelines=pipelines,
        ignore_scan_files=ignore_scan_files,
        ignore_sessions_files=ignore_session_scan_files,
        volume_atlas_selection=volume_atlas_selection,
        freesurfer_atlas_selection=freesurfer_atlas_selection,
        pvc_restriction=pvc_restriction,
        tsv_file=subjects_sessions_tsv,
        group_selection=group_selection,
        tracers_selection=pet_tracers_selection,
    )



from os import PathLike
from typing import Union


def initialize_tissues_spm_segment(
    parameters: dict, tissue_probability_map: Union[PathLike, None] = None
) -> tuple:
    """Prepare data structure for SPM segment interface

    Parameters
    ----------
    tissue_probability_map: PathLike, optional
        Path to the nifti tissue probability map. Default=None.

    Returns
    -------
    tuple
        "tissues" data structure for SPMSegment
    """

    from clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_utils import (
        get_tissue_tuples,
    )
    from clinica.utils.spm import get_tpm

    parameters["tissue_probability_maps"] = tissue_probability_map or get_tpm()

    tissue_tuples = get_tissue_tuples(
        parameters["tissue_probability_maps"],
        parameters["tissue_classes"],
        parameters["dartel_tissues"],
        parameters["save_warped_unmodulated"],
        parameters["save_warped_modulated"],
    )
    return tissue_tuples



def get_substitutions_datasink(bids_file, pipeline_name):
    suffix = "T1w" if pipeline_name == "t1-linear" else "flair"
    substitutions = [
        (
            f"{bids_file}Warped_cropped.nii.gz",
            f"{bids_file}_space-MNI152NLin2009cSym_desc-Crop_res-1x1x1_{suffix}.nii.gz",
        ),
        (
            f"{bids_file}0GenericAffine.mat",
            f"{bids_file}_space-MNI152NLin2009cSym_res-1x1x1_affine.mat",
        ),
        (
            f"{bids_file}Warped.nii.gz",
            f"{bids_file}_space-MNI152NLin2009cSym_res-1x1x1_{suffix}.nii.gz",
        ),
    ]
    return bids_file, substitutions


# Function used by the nipype interface.
# It crops an image based on the reference.
def crop_nifti(input_img, ref_crop):
    """Crop input image based on the reference.

    It uses nilearn `resample_to_img` function.

    Args:
        input_img (str): image to be processed
        ref_crop (str): template used to crop the image

    Returns:
       output_img (NIfTI image): crop image on disk.
       crop_template: (NIfTI image): output template on disk.
    """
    import os

    from nilearn.image import crop_img, resample_to_img

    basedir = os.getcwd()
    # crop_ref = crop_img(ref_img, rtol=0.5)
    # crop_ref.to_filename(os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz'))
    # crop_template = os.path.join(basedir, os.path.basename(input_img).split('.nii')[0] + '_cropped_template.nii.gz')

    # resample the individual MRI into the cropped template image
    crop_img = resample_to_img(input_img, ref_crop, force_resample=True)
    crop_img.to_filename(
        os.path.join(
            basedir, os.path.basename(input_img).split(".nii")[0] + "_cropped.nii.gz"
        )
    )

    output_img = os.path.join(
        basedir, os.path.basename(input_img).split(".nii")[0] + "_cropped.nii.gz"
    )
    crop_template = ref_crop

    return output_img, crop_template


def print_end_pipeline(anat, final_file):
    """Display end message for <subject_id> when <final_file> is connected."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_end_image

    print_end_image(get_subject_id(anat))



import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.t1_linear.t1_linear as pydra_t1_linear
from clinica.pipelines import cli_param
from clinica.pipelines.cli import cli as run_cli

pipeline_name = "pydra-t1-linear"


@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
def cli(
    bids_directory: str,
    caps_directory: str,
) -> None:
    """Affine registration of T1w images to the MNI standard space (Pydra engine)."""

    t1_linear_pipeline = pydra_t1_linear.build_core_workflow(
        name="t1-linear-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters={},
    )
    pydra_utils.run(t1_linear_pipeline)


run_cli.add_command(cli)

if __name__ == "__main__":
    cli()



from os import PathLike
from pathlib import Path
from typing import Dict, Optional

from ._inputs import (
    _build_thickness_array,
    _get_average_surface,
    _get_t1_freesurfer_custom_file_template,
    _read_and_check_tsv_file,
)
from ._model import create_glm_model


def clinica_surfstat(
    input_dir: PathLike,
    output_dir: PathLike,
    tsv_file: PathLike,
    design_matrix: str,
    contrast: str,
    glm_type: str,
    group_label: str,
    freesurfer_home: PathLike,
    surface_file: Optional[PathLike],
    feature_label: str,
    fwhm: Optional[int] = 20,
    threshold_uncorrected_pvalue: Optional[float] = 0.001,
    threshold_corrected_pvalue: Optional[float] = 0.05,
    cluster_threshold: Optional[float] = 0.001,
) -> None:
    """This function mimics the previous function `clinica_surfstat`
    written in MATLAB and relying on the MATLAB package SurfStat.

    This implementation is written in pure Python and rely on the
    package brainstat for GLM modeling.

    Results, both plots and matrices, will be written in the location
    specified through `output_dir`.

    The names of the output files within `output_dir` follow the
    conventions:

        <ROOT>_<SUFFIX>.<EXTENSION>

    EXTENSION can be:

        - "mat": for storing matrices (this is mainly for backward
          compatibility with the previous MATLAB implementation of
          this function).
        - "json": for storing matrices in a more Pythonic way.
        - "png" for surface figures.

    SUFFIX can be:

        - "coefficients": relative to the model's beta coefficients.
        - "TStatistics": relative to the T-statistics.
        - "uncorrectedPValue": relative to the uncorrected P-values.
        - "correctedPValues": relative to the corrected P-values.
        - "FDR": Relative to the False Discovery Rate.

    ROOT can be:

        - For group comparison GLM with an interaction term:

            interaction-<contrast_name>_measure-<feature_label>_fwhm-<fwhm>

        - For group comparison GLM without an interaction term:

            group-<group_label>_<contrast_name>_measure-<feature_label>_fwhm-<fwhm>

        - For correlation GLM:

            group-<group_label>_correlation-<contrast_name>-<contrast_sign>_measure-<feature_label>_fwhm-<fwhm>

    Parameters
    ----------
    input_dir : PathLike
        Path to the input folder.

    output_dir : PathLike
        Path to the output folder for storing results.

    tsv_file : PathLike
        Path to the TSV file `subjects.tsv` which contains the
        necessary metadata to run the statistical analysis.

        .. warning::
            The column names need to be accurate because they
            are used to defined contrast and model terms.
            Please double check for typos.

    design_matrix : str
        The design matrix specified in string format.
        For example "1 + Label"

    contrast : str
        The contrast to be used in the GLM, specified in string format.

        .. warning::
            The contrast needs to be in the design matrix.

    glm_type : {"group_comparison", "correlation"}
        Type of GLM to run:
            - "group_comparison": Performs group comparison.
              For example "AD - ND".
            - "correlation": Performs correlation analysis.

    group_label : str
        The label for the group. This is used in the output file names
        (see main description of the function).

    freesurfer_home : PathLike
        The path to the home folder of Freesurfer.
        This is required to get the fsaverage templates.

    surface_file : PathLike, optional
        The path to the surface file to analyze.
        Typically the cortical thickness.
        If `None`, the surface file will be the t1 freesurfer template.

    feature_label : str
        The label used for the measure. This is used in the output file
        names (see main description of the function).

    fwhm : int, optional
        The smoothing FWHM. This is used in the output file names.
        Default=20.

    threshold_uncorrected_pvalue : float, optional
        The threshold to be used with uncorrected P-values. Default=0.001.

    threshold_corrected_pvalue : float, optional
        The threshold to be used with corrected P-values. Default=0.05.

    cluster_threshold : float, optional
        The threshold to be used to declare clusters as significant. Default=0.05.
    """
    # Load subjects data
    df_subjects = _read_and_check_tsv_file(tsv_file)
    if surface_file is None:
        surface_file = _get_t1_freesurfer_custom_file_template(input_dir)
    thickness = _build_thickness_array(input_dir, surface_file, df_subjects, fwhm)

    # Load average surface template
    fsaverage_path = freesurfer_home / Path("subjects/fsaverage/surf")
    average_surface, average_mesh = _get_average_surface(fsaverage_path)

    # Build and run GLM model
    glm_model = create_glm_model(
        glm_type,
        design_matrix,
        df_subjects,
        contrast,
        feature_label,
        group_label=group_label,
        fwhm=fwhm,
        threshold_uncorrected_pvalue=threshold_uncorrected_pvalue,
        threshold_corrected_pvalue=threshold_corrected_pvalue,
        cluster_threshold=cluster_threshold,
    )
    glm_model.fit(thickness, average_surface)
    glm_model.save_results(output_dir, ["json", "mat"])
    glm_model.plot_results(output_dir, ["nilearn_plot_surf_stat_map"], average_mesh)



from click_option_group import optgroup

option = optgroup.option

pipeline_specific_options = optgroup.group(
    "Pipeline-specific options", help="Options specific to the pipeline being run"
)

common_pipelines_options = optgroup.group(
    "Common pipelines options", help="Options common to all Clinica pipelines"
)

advanced_pipeline_options = optgroup.group(
    "Advanced pipeline options", help="For experts only"
)


def custom_pipeline_options(name: str, *args, **kwargs):
    return optgroup.group(name=name, *args, **kwargs)



"""Convert the UKB dataset into BIDS."""

from os import PathLike


def convert_images(
    path_to_dataset: PathLike,
    bids_dir: PathLike,
    path_to_clinical: PathLike,
):
    """Convert the entire dataset to BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """

    import clinica.iotools.bids_utils as bids

    from .ukb_utils import (
        complete_clinical,
        dataset_to_bids,
        find_clinical_data,
        intersect_data,
        read_imaging_data,
        write_bids,
    )

    # read the clinical data files
    df_clinical = find_clinical_data(path_to_clinical)

    # makes a df of the imaging data
    imaging_data = read_imaging_data(path_to_dataset)

    # intersect the data
    df_clinical = intersect_data(imaging_data, df_clinical)

    # complete clinical data
    df_clinical = complete_clinical(df_clinical)

    # build the tsv
    result = dataset_to_bids(df_clinical)

    write_bids(
        to=bids_dir,
        participants=result["participants"],
        sessions=result["sessions"],
        scans=result["scans"],
        dataset_directory=path_to_dataset,
    )
    readme_data = {
        "link": "https://www.ukbiobank.ac.uk/",
        "desc": (
            "UK Biobank is a large-scale biomedical database and research resource, containing in-depth genetic and "
            "health information from half a million UK participants. The database is regularly augmented with "
            "additional data and is globally accessible to approved researchers undertaking vital research into the "
            "most common and life-threatening diseases. It is a major contributor to the advancement of modern "
            "medicine it and has led to the discovery of several scientific advances and numerous treatments to "
            "improve human health."
        ),
    }
    bids.write_modality_agnostic_files(
        study_name="UKB", readme_data=readme_data, bids_dir=bids_dir
    )



import numpy as np

from clinica.pipelines.machine_learning import algorithm, base, input, validation


class VoxelBasedKFoldDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        output_dir,
        fwhm=0,
        modulated="on",
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        mask_zeros=True,
        n_threads=15,
        n_folds=10,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(VoxelBasedKFoldDualSVM, self).__init__(
            input.CAPSVoxelBasedInput,
            validation.KFoldCV,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class VoxelBasedRepKFoldDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        output_dir,
        fwhm=0,
        modulated="on",
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        mask_zeros=True,
        n_threads=15,
        n_iterations=100,
        n_folds=10,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(VoxelBasedRepKFoldDualSVM, self).__init__(
            input.CAPSVoxelBasedInput,
            validation.RepeatedKFoldCV,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class VoxelBasedRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        output_dir,
        fwhm=0,
        modulated="on",
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        mask_zeros=True,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super().__init__(
            input.CAPSVoxelBasedInput,
            validation.RepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class VertexBasedRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        output_dir,
        image_type="PET",
        acq_label=None,
        suvr_reference_region=None,
        fwhm=20,
        precomputed_kernel=None,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-10, 2, 1000),
        splits_indices=None,
    ):

        super(VertexBasedRepHoldOutDualSVM, self).__init__(
            input.CAPSVertexBasedInput,
            validation.RepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class RegionBasedRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(RegionBasedRepHoldOutDualSVM, self).__init__(
            input.CAPSRegionBasedInput,
            validation.RepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class RegionBasedRepHoldOutLogisticRegression(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(RegionBasedRepHoldOutLogisticRegression, self).__init__(
            input.CAPSRegionBasedInput,
            validation.RepeatedHoldOut,
            algorithm.LogisticReg,
            locals(),
            output_dir,
        )


class RegionBasedRepHoldOutRandomForest(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        n_estimators_range=(100, 200, 400),
        max_depth_range=[None],
        min_samples_split_range=[2],
        max_features_range=("auto", 0.25, 0.5),
        splits_indices=None,
    ):

        super(RegionBasedRepHoldOutRandomForest, self).__init__(
            input.CAPSRegionBasedInput,
            validation.RepeatedHoldOut,
            algorithm.RandomForest,
            locals(),
            output_dir,
        )


class RegionBasedLearningCurveRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        n_learning_points=10,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
    ):

        super(RegionBasedLearningCurveRepHoldOutDualSVM, self).__init__(
            input.CAPSRegionBasedInput,
            validation.LearningCurveRepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class VoxelBasedLearningCurveRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        output_dir,
        fwhm=0,
        modulated="on",
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        mask_zeros=True,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        n_learning_points=10,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
    ):

        super(VoxelBasedLearningCurveRepHoldOutDualSVM, self).__init__(
            input.CAPSVoxelBasedInput,
            validation.LearningCurveRepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class RegionBasedRepKFoldDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        n_folds=10,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(RegionBasedRepKFoldDualSVM, self).__init__(
            input.CAPSRegionBasedInput,
            validation.RepeatedKFoldCV,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class CAPSTsvRepHoldOutDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        dataset,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(CAPSTsvRepHoldOutDualSVM, self).__init__(
            input.CAPSTSVBasedInput,
            validation.RepeatedHoldOut,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


class CAPSTsvRepHoldOutRandomForest(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        atlas,
        dataset,
        output_dir,
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        n_threads=15,
        n_iterations=100,
        test_size=0.3,
        grid_search_folds=10,
        balanced=True,
        n_estimators_range=(100, 200, 400),
        max_depth_range=[None],
        min_samples_split_range=[2],
        max_features_range=("auto", 0.25, 0.5),
        splits_indices=None,
    ):

        super(CAPSTsvRepHoldOutRandomForest, self).__init__(
            input.CAPSTSVBasedInput,
            validation.RepeatedHoldOut,
            algorithm.RandomForest,
            locals(),
            output_dir,
        )


# SVM reg


class VoxelBasedREGRepKFoldDualSVM(base.MLWorkflow):
    def __init__(
        self,
        caps_directory,
        subjects_visits_tsv,
        diagnoses_tsv,
        group_label,
        image_type,
        output_dir,
        fwhm=0,
        modulated="on",
        acq_label=None,
        suvr_reference_region=None,
        use_pvc_data=False,
        precomputed_kernel=None,
        mask_zeros=True,
        n_threads=15,
        n_iterations=100,
        n_folds=10,
        test_size=0.1,
        grid_search_folds=10,
        balanced=True,
        c_range=np.logspace(-6, 2, 17),
        splits_indices=None,
    ):

        super(VoxelBasedREGRepKFoldDualSVM, self).__init__(
            input.CAPSTSVBasedInput,
            validation.RepeatedKFoldCV,
            algorithm.DualSVMAlgorithm,
            locals(),
            output_dir,
        )


# TSV


class TsvRepHoldOutRandomForest(base.MLWorkflow):
    def __init__(
        self,
        data_tsv,
        columns,
        output_dir,
        n_threads=20,
        n_iterations=250,
        test_size=0.2,
        grid_search_folds=10,
        balanced=True,
        n_estimators_range=(100, 200, 400),
        max_depth_range=[None],
        min_samples_split_range=[2],
        max_features_range=("auto", 0.25, 0.5),
        splits_indices=None,
        inner_cv=False,
    ):

        super(TsvRepHoldOutRandomForest, self).__init__(
            input.TsvInput,
            validation.RepeatedHoldOut,
            algorithm.RandomForest,
            locals(),
            output_dir,
        )



def init_input_node(pet_nii):
    import nibabel as nib

    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    # Extract image ID
    image_id = get_subject_id(pet_nii)

    # Check that the PET file is a 3D volume
    img = nib.load(pet_nii)
    if len(img.shape) == 4:
        error_msg = (
            f"Clinica does not handle 4D volumes for {image_id.replace('_', ' | ')}"
        )
        cprint(error_msg, lvl="error")
        raise NotImplementedError(error_msg)

    # Print begin message
    print_begin_image(image_id)

    return pet_nii


def _check_non_empty_tissue_list(tissues: list) -> None:
    """Check that provided list is non-empty."""
    if len(tissues) == 0:
        raise RuntimeError(
            "The length of the list of tissues must be greater than zero."
        )


def _load_tissues(tissues: list):
    """Aggregates the image data contained in the tissue images provided.

    Parameters
    ----------
    tissues : list
        List of tissue images to aggregate.

    Returns
    -------
    data : np.ndarray
        Aggregated data.

    affine : np.ndarray
        Affine of the first image, acting as the affine
        of the aggregated image under the assumption that
        all images have the same affine.

    header : Nifti1Header
        Header of the aggregated image.
    """
    import nibabel as nib
    import numpy as np

    from clinica.pipelines.pet_volume.pet_volume_utils import (  # noqa
        _check_non_empty_tissue_list,
    )

    _check_non_empty_tissue_list(tissues)
    img_0 = nib.load(tissues[0])
    shape = list(img_0.get_fdata(dtype="float32").shape)
    data = np.zeros(shape=shape)
    for image in tissues:
        data += nib.load(image).get_fdata(dtype="float32")
    return data, img_0.affine, img_0.header


def create_binary_mask(
    tissues: list,
    threshold: float = 0.3,
) -> str:
    """Create a binary mask Nifti1Image from the list of tissues.

    Tissue images are summed and the result is thresholded with the
    provided `threshold` input.

    Parameters
    ----------
    tissues : list
        List of paths to tissue Nifti1Images. Must be non-empty.

    threshold : float, optional
        Threshold to apply when binarizing the Nifti1Image.
        Default=0.3.

    Returns
    -------
    out_mask : str
        Path to the binary mask Nifti1Image as a string.
    """
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib

    from clinica.pipelines.pet_volume.pet_volume_utils import _load_tissues  # noqa

    data, affine, header = _load_tissues(tissues)
    data = (data > threshold) * 1.0
    out_mask = join(getcwd(), basename(tissues[0]) + "_brainmask.nii")
    mask = nib.Nifti1Image(data, affine, header=header)
    nib.save(mask, out_mask)
    return out_mask


def apply_binary_mask(image: str, binary_mask: str) -> str:
    """Apply the provided `binary_mask` to the provided `image`.

    Parameters
    ----------
    image : str
        Path to the Nifti1Image to apply the mask on.

    binary_mask : str
        Path to the Nifti1Image containing the mask.

    Returns
    -------
    masked_image_path : str
        Path to the masked Nifti1Image.
    """
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib

    original_image = nib.load(image)
    mask = nib.load(binary_mask)

    data = original_image.get_fdata(dtype="float32") * mask.get_fdata(dtype="float32")

    masked_image_path = join(getcwd(), "masked_" + basename(image))
    masked_image = nib.Nifti1Image(
        data, original_image.affine, header=original_image.header
    )
    nib.save(masked_image, masked_image_path)
    return masked_image_path


def create_pvc_mask(tissues: list) -> str:
    """Create a pvc mask from tissue list.

    Parameters
    ----------
    tissues : list
        List of paths to tissue Nifti1Images. Must be non-empty.

    Returns
    -------
    out_mask : str
        Path to the resulting mask Nifti1Image.
    """
    from os import getcwd
    from os.path import join

    import nibabel as nib
    import numpy as np

    from clinica.pipelines.pet_volume.pet_volume_utils import _load_tissues  # noqa

    background, affine, header = _load_tissues(tissues)
    shape = background.shape
    shape += tuple([len(tissues) + 1])
    data = np.empty(shape=shape, dtype=np.float64)
    for i, tissue in enumerate(tissues):
        image = nib.load(tissue)
        data[..., i] = np.array(image.get_fdata(dtype="float32"))
    background = 1.0 - background
    data[..., len(tissues)] = np.array(background)
    out_mask = join(getcwd(), "pvc_mask.nii")
    mask = nib.Nifti1Image(data, affine, header=header)
    nib.save(mask, out_mask)
    return out_mask


def pet_pvc_name(pet_image: str, pvc_method: str) -> str:
    """Build the name for the PET PVC interface.

    Parameters
    ----------
    pet_image : str
        Path to the PET scan file.

    pvc_method : str
        Name of the PVC method. This will be concatenated
        with the `pet_image` filename.

    Returns
    -------
    pet_pvc_path : str
        Name for the PET PVC interface.

    Examples
    --------
    >>> pet_pvc_name(
    ...     "/home/bids/sub-01/ses-M00/pet/sub-01_ses-M00_task-rest_trc-av45_pet.nii.gz",
    ...     "RBV"
    ...)
    'pvc-rbv_sub-01_ses-M00_task-rest_trc-av45_pet.nii.gz'

    """
    from os.path import basename

    return "pvc-" + pvc_method.lower() + "_" + basename(pet_image)


def normalize_to_reference(pet_image: str, region_mask: str) -> str:
    """Normalize the provided `pet_image` by dividing by the mean
    value of the region defined by the provided `region_mask`.

    Parameters
    ----------
    pet_image : str
        Path to the Nifti1Image which should be normalized.

    region_mask : str
        Path to the mask to be used to define the region.

    Returns
    -------
    suvr_pet_path : str
        Path to the normalized Nifti1Image.
    """
    from os import getcwd
    from os.path import basename, join

    import nibabel as nib
    import numpy as np

    pet = nib.load(pet_image)
    ref = nib.load(region_mask)

    region = pet.get_fdata(dtype="float32") * ref.get_fdata(dtype="float32")
    region_mean = np.nanmean(np.where(region != 0, region, np.nan))

    data = pet.get_fdata(dtype="float32") / region_mean

    suvr_pet_path = join(getcwd(), "suvr_" + basename(pet_image))

    suvr_pet = nib.Nifti1Image(data, pet.affine, header=pet.header)
    nib.save(suvr_pet, suvr_pet_path)

    return suvr_pet_path


def atlas_statistics(in_image: str, in_atlas_list: list) -> list:
    """Generate regional measure from atlas_list in TSV files.

    For each atlas name provided it calculates for the input image the mean
    for each region in the atlas and saves it to a TSV file.

    Parameters
    ----------
    in_image : str
        Path to the Nifti image.

    in_atlas_list : List
        List of names of atlas to be applied.

    Returns
    -------
    atlas_statistics : List
        List of paths to TSV files.
    """
    from os import getcwd
    from os.path import abspath, join

    from nipype.utils.filemanip import split_filename

    from clinica.utils.atlas import AtlasAbstract
    from clinica.utils.statistics import statistics_on_atlas

    orig_dir, base, ext = split_filename(str(in_image))
    atlas_classes = AtlasAbstract.__subclasses__()
    atlas_statistics_list = []
    for atlas in in_atlas_list:
        for atlas_class in atlas_classes:
            if atlas_class.get_name_atlas() == atlas:
                out_atlas_statistics = abspath(
                    join(getcwd(), base + "_space-" + atlas + "_statistics.tsv")
                )
                statistics_on_atlas(str(in_image), atlas_class(), out_atlas_statistics)
                atlas_statistics_list.append(out_atlas_statistics)
                break
    return atlas_statistics_list


def get_from_list(in_list, index):
    return in_list[index]



# from pathlib import PurePath
import abc
from os import PathLike
from typing import Dict, List

import nipype.interfaces.io as nio
import pydra
from nipype.interfaces.base import Directory, DynamicTraitedSpec, Str, isdefined, traits
from nipype.interfaces.io import IOBase, add_traits
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pydra.query import BIDSQuery, CAPSFileQuery, CAPSGroupQuery, CAPSQuery


class CAPSDataGrabberInputSpec(DynamicTraitedSpec):
    base_dir = Directory(exists=True, desc="Path to CAPS Directory.", mandatory=True)
    output_query = traits.Dict(key_trait=Str, desc="Queries for outfield outputs")
    raise_on_empty = traits.Bool(
        True,
        usedefault=True,
        desc="Generate exception if list is empty for a given field",
    )


class CAPSDataGrabber(IOBase):
    input_spec = CAPSDataGrabberInputSpec
    output_spec = DynamicTraitedSpec

    def __init__(self, **kwargs):
        super(CAPSDataGrabber, self).__init__(**kwargs)

        if not isdefined(self.inputs.output_query):
            self.inputs.output_query = {}

        # used for mandatory inputs check
        undefined_traits = {}
        self.inputs.trait_set(trait_change_notify=False, **undefined_traits)
        self.sessions = None
        self.subjects = None

    def _list_outputs(self):
        from clinica.utils.participant import get_subject_session_list

        sessions, subjects = get_subject_session_list(
            self.inputs.base_dir,
            is_bids_dir=False,
        )
        self.sessions = sessions
        self.subjects = subjects

        output_query = {}
        for k, query in self.inputs.output_query.items():
            if isinstance(query, list):
                temp = [self._execute_single_query(q) for q in query]
                if len(temp) != len(self.subjects) and len(temp[0]) == len(
                    self.subjects
                ):
                    temp = [list(i) for i in zip(*temp)]  # Transpose
                output_query[k] = temp
            else:
                output_query[k] = self._execute_single_query(query)
        return output_query

    @abc.abstractmethod
    def _execute_single_query(self, query: dict) -> list:
        pass

    def _add_output_traits(self, base):
        return add_traits(base, list(self.inputs.output_query.keys()))


class CAPSFileDataGrabber(CAPSDataGrabber):
    def _execute_single_query(self, query: Dict) -> List[str]:
        from clinica.utils.inputs import clinica_file_reader

        return clinica_file_reader(
            self.subjects,
            self.sessions,
            self.inputs.base_dir,
            query,
        )[0]


class CAPSGroupDataGrabber(CAPSDataGrabber):
    def _execute_single_query(self, query: Dict) -> List[str]:
        from clinica.utils.inputs import clinica_group_reader

        return clinica_group_reader(self.inputs.base_dir, query)


@pydra.mark.task
@pydra.mark.annotate({"return": {"output_file": str}})
def bids_writer(output_dir: PathLike, output_file: PathLike) -> str:
    """
    (toy) Interface to echo bids files

    Parameters
    ----------
    output_dir : PathLike
        output directory
    output_file : PathLike
        The output file to write
    Returns
    -------
    An BIDS reader based workflow
    """

    print("printing core vars in output_workflow", output_file)
    print("should go to location", output_dir)

    return output_file


def bids_reader(query: BIDSQuery, input_dir: PathLike):
    """
    Parameters
    ----------
    query : BIDSQuery
        BIDS dataset reader query.
    input_dir :  PathLike
        The BIDS input directory.

    Returns
    -------
    Nipype1Task
        The task used for reading files from BIDS.
    """
    from pydra.tasks.bids import BIDSDatasetReader

    return BIDSDatasetReader(output_query=query.query).to_task(
        name="bids_reader_task", dataset_path=input_dir
    )


def caps_reader(query: CAPSQuery, input_dir: PathLike):
    """
    Parameters
    ----------
    query : CAPSQuery
        Input to CAPSDataGrabber.
    input_dir :  PathLike
        The CAPS input directory.

    Returns
    -------
    Nipype1Task
        The task used for reading files from CAPS.
    """
    if isinstance(query, CAPSFileQuery):
        grabber = CAPSFileDataGrabber
        name = "caps_file_reader_task"
    elif isinstance(query, CAPSGroupQuery):
        grabber = CAPSGroupDataGrabber
        name = "caps_group_reader_task"
    else:
        raise TypeError(
            f"caps_reader received an unexpected query type {type(query)}. "
            "Supported types are: CAPSFileQuery and CAPSGroupQuery."
        )
    caps_data_grabber = grabber(output_query=query.query)
    caps_reader_task = Nipype1Task(
        name=name,
        interface=caps_data_grabber,
        base_dir=input_dir,
        output_query=query.query,
    )
    return caps_reader_task



"""This module handles Clinica exceptions."""


class ClinicaException(Exception):
    """Base class for Clinica exceptions."""


class ClinicaMissingDependencyError(ClinicaException):
    """Base class for Clinica dependencies errors."""


class ClinicaBIDSError(ClinicaException):
    """Base class for BIDS errors."""


class ClinicaCAPSError(ClinicaException):
    """Base class for CAPS errors."""


class ClinicaParserError(ClinicaException):
    """Base class for parser errors."""


class ClinicaInconsistentDatasetError(ClinicaException):
    """Base class for inconsistent datasets errors."""

    def __init__(self, cross_subj: list):
        max_subjects_displayed = 50
        bound = min(len(cross_subj), max_subjects_displayed)
        self.msg = (
            f"{len(cross_subj)} subjects in your BIDS folder did not respect the longitudinal organisation "
            f"from BIDS specification.\nThe subjects concerned are (showing only the first {bound}):\n\t- "
        )
        self.msg += "\n\t- ".join(cross_subj[:bound])
        self.msg += (
            f"\nClinica does not know how to handle cross sectional dataset, but it "
            "can convert it to a Clinica compliant form (using session ses-M00)"
        )
        super(ClinicaException, self).__init__(self.msg)
